# coding: UTF-8
"""
Copyright (c) 2009 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk

from .spell_entry import SpellEntry


class SearchBar(gtk.Table):

	__gtype_name__ = "SearchBar"


	@property
	def textview_callback(self): return self._textview_callback
	@textview_callback.setter
	def textview_callback(self, callback):
		self._textview_callback = callback


	@property
	def textview(self):
		if self.textview_callback:
			return self.textview_callback()
		return self._textview
	@textview.setter
	def textview(self, textview):
		if self.textview_callback:
			self._textview = self.textview_callback()
		else:
			self._textview = textview


	@property
	def search_term(self): return self.search_entry.get_text()
	@search_term.setter
	def search_term(self, text): self.search_entry.set_text(text)


	@property
	def autohide(self): return self._autohide
	@autohide.setter
	def autohide(self, value):
		self._autohide = value

		# connect/disconnect the focus-out-event for auto-hiding
		if value:
			id = self.search_entry.connect(
					"focus-out-event",
					self.search_entry_focus_out_cb)
			self._autohide_sig_id = id
		else:
			if hasattr(self, "_autohide_sig_id"):
				self.search_entry.disconnect(self._autohide_sig_id)


	@property
	def default_buttons(self): return self._default_buttons
	@default_buttons.setter
	def default_buttons(self, value):
		""" default button behaviour (buttons will be connected
			to self.search_button_clicked_cb) if value == True
			Otherwise the handler will be disconnected.
		"""
		if value:
			h1 = self.search_button.connect("clicked",
					self.search_button_clicked_cb)

			h2 = self.search_entry.connect("activate",
					self.search_button_clicked_cb)
			self._btn_handler_1 = h1
			self._btn_handler_2 = h2

		else:
			if hasattr(self, "_btn_handler_1"):
				self.disconnect(self._btn_handler_1)
				self.disconnect(self._btn_handler_2)


	def __init__(self, textview=None, textview_callback = None, autohide=True, default_buttons=True):
		"""
			textview = TextView to operate on
			textview_callback = callback wich returns a TextView to operate on
			autohide = hide on focus loss if enabled (True)
		"""

		super(SearchBar,self).__init__(rows=1, columns=2)


		# arange widgets

		self.set_property("row-spacing", 1)

		self.search_entry = SpellEntry()

		self.attach(self.search_entry, 0, 1, 1, 2)
		self.child_set_property(self.search_entry, "y-options", gtk.SHRINK)

		self.search_button = gtk.ToolButton(stock_id = gtk.STOCK_FIND)
		self.attach(self.search_button, 1, 2, 1, 2)

		self.child_set_property(self.search_button,
								"y-options",
								gtk.SHRINK)
		self.child_set_property(self.search_button,
								"x-options",
								gtk.SHRINK)

		# setup own properties

		self.last_iter = None
		self.last_result = ""

		self.textview_callback = textview_callback
		self.textview = textview
		self.autohide = autohide

		# connect widget signals

		self.default_buttons = default_buttons


	def search_entry_focus_out_cb(self, entry, event):
		self.hide()


	def search_further(self):
		self.search_button_clicked_cb(None)


	def search_button_clicked_cb(self, button):

		if not self.search_term or not self.textview:
			return

		if not self.last_iter or self.last_result != self.search_term:
			self.last_iter = self.textview.get_buffer().get_start_iter()

		result = self.last_iter.forward_search(self.search_term,
										gtk.TEXT_SEARCH_TEXT_ONLY)

		if not result:
			return

		self.last_iter = result[1]
		self.last_result = self.textview.get_buffer().get_text(*result)

		self.textview.get_buffer().select_range(*result)

		# scroll the textview
		self.textview.scroll_to_iter(result[0], 0.0)


	def grab_focus(self):
		self.search_entry.grab_focus()



def test():
	win = gtk.Window()
	win.resize(400,400)

	vbox = gtk.VBox()
	tv = gtk.TextView()
	tv.get_buffer().set_text(
"""Oh hai!
Das ist ein Test!
Wer hier suchen will der findet das.
Auch doppelte und doppelte Woerter.
""")
	bar = SearchBar(textview_callback=lambda: tv, autohide=False)

	vbox.add(tv)
	vbox.add(bar)
	vbox.child_set_property(bar, "fill", False)
	vbox.child_set_property(bar, "expand", False)

	bar.grab_focus()

	win.add(vbox)
	win.connect("destroy", lambda *x: gtk.main_quit())

	win.show_all()

	gtk.main()


if __name__ == "__main__":
	test()




# coding: UTF-8
"""
Copyright (c) 2008 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import logging
import webbrowser

from .. import config

eventIDs = {}

def _resetCursor(widget, event, window, cursor):
	window.set_cursor(cursor)
	widget.disconnect(eventIDs[widget])
	del eventIDs[widget]

def _build_url_menu(browser, url):
	menu = gtk.Menu()
	cb = gtk.Clipboard()

	if browser:
		openitem = gtk.MenuItem(label="Open")
		openitem.connect("activate",
			lambda w,b: b.open(url), browser)

		menu.append(openitem)

	copyitem = gtk.MenuItem(label="Copy URL")
	copyitem.connect("activate",
		lambda w,u,c: c.set_text(u), url, cb)
	menu.append(copyitem)

	return menu

def _get_browser():
	name = config.get("tekka","browser")
	try:
		if name and webbrowser.get(name):
			browser = webbrowser.get(name)
		else:
			browser = webbrowser
	except webbrowser.Error:
		logging.error("Could not open a browser")
		browser = None

	except TypeError:
		logging.debug("Fetching bug in python2.4")
		browser = None
	return browser

def URLHandler(texttag, widget, event, iter, url):
	""" do URL specific stuff """

	if event.type == gtk.gdk.MOTION_NOTIFY:
		# TODO: LP#303346
		# TODO:: if the state is BUTTON1_MASK and this event
		# TODO:: occurs, create a object aware of DND and
		# TODO:: let it take over.
		if event.state & gtk.gdk.BUTTON1_MASK:
			return False

		# cursor moved on the URL, change cursor to HAND2
		cursor = gtk.gdk.Cursor(gtk.gdk.HAND2)
		textWin = widget.get_window(gtk.TEXT_WINDOW_TEXT)
		textWin.set_cursor(cursor)

		# add signal to reset the cursor
		if not eventIDs.has_key(widget):
			id = widget.connect("motion-notify-event", _resetCursor,
					textWin, gtk.gdk.Cursor(gtk.gdk.XTERM))
			eventIDs[widget] = id

		return True

	browser = _get_browser()

	if event.type == gtk.gdk.BUTTON_RELEASE:

		if (event.button == 1 and browser
		and not widget.get_buffer().get_has_selection()):
			# open URL in browser
			browser.open(url)

	elif event.type == gtk.gdk.BUTTON_PRESS:

		if event.button == 3:
			# print menu for URL actions
			menu = _build_url_menu(browser, url)

			menu.show_all()
			menu.popup(None, None, None, button=event.button,
				activate_time=event.time)

			return True




"""
Copyright (c) 2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gobject
import gtk


class FadingBox(gtk.EventBox):

	""" Has the ability to fade it's background color to another """

	""" NOTE: color attributes suffer from integer overflow (0xffff) """

	MAX_FADING_TIME = 700 # miliseconds

	# background color as defined in style rc
	origin_bg_color = property(lambda s: s._origin_bg_color.copy(), None)

	# bool property if fading is in progress
	is_fading = property(lambda s: s._timeout_id != None, None)

	def __init__(self):
		super(FadingBox,self).__init__()

		self._origin_bg_color = self.get_style().bg[gtk.STATE_NORMAL].copy()
		self._timeout_id = None
		self._timer = None


	def reset_background(self):
		""" reset the background color to what it should be according
			to the gtkrc loaded
		"""

		style = self.get_style().copy()
		style.bg[gtk.STATE_NORMAL] = self._origin_bg_color.copy()
		self.set_style(style)


	def _fade_timeout(self, color):

		def do_timeout(self, color):
			style = self.get_style().copy()
			style.bg[gtk.STATE_NORMAL] = color.copy()
			self.set_style(style)

			self.stop_fading()
		gobject.idle_add(do_timeout, self, color)
		return False


	def _fade_channel(self, channel, reference):

		if channel == reference:
			return (True, reference)

		if channel > reference:
			up = (channel / 10) + 1
			if channel-up < 0 or channel-up < reference:
				return (True, reference)
			return (False, channel-up)

		else:
			up = (reference / 10) + 1
			if channel+up >= 0xffff or channel+up > reference:
				return (True, reference)
			return (False, channel+up)


	def _fade_bg(self, color):
		""" modifiy each color channel of the background color according
			to color and refresh the style with the new background color
		"""

		if not self._timeout_id:
			return False

		bg_color = self.bg_color

		(rdone, bg_color.red) = self._fade_channel(bg_color.red,
												   color.red)
		(gdone, bg_color.green) = self._fade_channel(bg_color.green,
													 color.green)
		(bdone, bg_color.blue) = self._fade_channel(bg_color.blue,
													color.blue)


		self.bg_color = bg_color

		style = self.get_style().copy()
		style.bg[gtk.STATE_NORMAL] = bg_color.copy()
		self.set_style(style)

		if rdone and gdone and bdone:
			self.stop_fading()
			return False

		return True


	def fade(self, to_color, interval=40):
		""" fade the background color of this box to the given color.
			interval is the time in miliseconds between each fade
		"""

		if self._timeout_id:
			return False # already fading


		style = self.get_style()
		self.bg_color = style.bg[gtk.STATE_NORMAL].copy()

		self._timeout_id = gobject.timeout_add(interval,
											   self._fade_bg,
											   to_color)

		self._timer = gobject.timeout_add(self.MAX_FADING_TIME,
										  self._fade_timeout,
										  to_color)

		return True


	def stop_fading(self):

		if self._timeout_id:
			gobject.source_remove(self._timeout_id)
			gobject.source_remove(self._timer)

			self._timeout_id = None
			self._timer = None

			self.emit("fade-finished")


gobject.signal_new("fade-finished", FadingBox, gobject.SIGNAL_ACTION,
				   None, ())


if __name__ == "__main__":

	win = gtk.Window()
	win.set_default_size(500,300)

	vbox = gtk.VBox()
	win.add(vbox)

	box = FadingBox()
	vbox.pack_start(box)

	btn = gtk.Button("Green")
	vbox.add(btn)

	btn2 = gtk.Button("Background")
	vbox.add(btn2)

	btn3 = gtk.Button("Blue")
	vbox.add(btn3)

	btn4 = gtk.Button("Red")
	vbox.add(btn4)

	box.fade(gtk.gdk.Color("#0000ff"))

	btn5 = gtk.Button("Black")
	vbox.add(btn5)

	def btn_clicked_cb(btn):
		"""
		def do_setup():

			def new_fade(box):
				if new_fade.ran:
					return

				print "new fade"
				c = gtk.gdk.Color("#FF0000")
				box.fade(c)
				new_fade.ran = True

			c = gtk.gdk.Color("#00FF00")
			box.fade(c)
			new_fade.ran = False

			if not btn_clicked_cb.init:
				box.connect("fade-finished", new_fade)
				btn_clicked_cb.init = True


		box.stop_fading()
		box.reset_background()

		gobject.timeout_add(1000, do_setup)
		"""

		"""
		def finished_cb(box, callback):
			self = finished_cb

			if self.count == 0:
				callback()
				return

			if self.back_fade:
				gobject.idle_add(box.fade, self.two)
			else:
				gobject.idle_add(box.fade, self.one)

			self.back_fade = not self.back_fade
			self.count -= 1

		def killsig():
			box.disconnect(killsig.handle)

		handle = box.connect("fade-finished", finished_cb, killsig)
		finished_cb.back_fade = False
		finished_cb.one = gtk.gdk.Color("#0f0")
		finished_cb.two = box.origin_bg_color
		finished_cb.count = 4

		killsig.handle = handle

		box.fade(gtk.gdk.Color("#0f0"))
		"""

		box.fade(gtk.gdk.Color("#0f0"))


	def btn2_clicked_cb(btn):

		box.fade(box.origin_bg_color)


	def btn3_clicked_cb(btn):

		box.fade(gtk.gdk.Color("#00f"))


	def btn4_clicked_cb(btn):

		box.fade(gtk.gdk.Color("#f00"))


	def btn5_clicked_cb(btn):

		box.fade(gtk.gdk.Color("#000"))


	btn_clicked_cb.init = False
	btn.connect("clicked", btn_clicked_cb)
	btn2.connect("clicked", btn2_clicked_cb)
	btn3.connect("clicked", btn3_clicked_cb)
	btn4.connect("clicked", btn4_clicked_cb)
	btn5.connect("clicked", btn5_clicked_cb)

	win.connect("destroy", lambda w: gtk.main_quit())
	win.show_all()

	gtk.main()



"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
from gettext import gettext as _

from .. import com
from .. import config

from .output_window import OutputWindow

class WelcomeWindow(OutputWindow):

	def __init__(self, *args, **kwargs):
		OutputWindow.__init__(self, *args, **kwargs)

		self.set_properties(
			hscrollbar_policy=gtk.POLICY_AUTOMATIC,
			vscrollbar_policy = gtk.POLICY_NEVER )

		self.remove(self.textview)

		self.table = gtk.Table(rows = 2, columns = 2)
		self.table.set_homogeneous(False)
		self.table.set_property("border-width", 12)

		try:
			self.pixbuf = gtk.icon_theme_get_default().load_icon("tekka",128,0)
		except:
			self.pixbuf = None

		self.image = gtk.image_new_from_pixbuf(self.pixbuf)

		# Create Header label
		self.label = gtk.Label()
		self.label.set_property("yalign", 1)
		self.label.set_property("xalign", 0.05)
		self.label.set_markup(
			_("<big><b>Welcome to tekka!</b></big>"))

		# Add Image to table
		self.ibox = gtk.EventBox()
		self.ibox.add(self.image)

		self.table.attach(self.ibox, 0, 1, 0, 2,
			xoptions=gtk.FILL|gtk.SHRINK,
			yoptions = gtk.FILL|gtk.EXPAND)

		# Add Label to table
		self.lbox = gtk.EventBox()
		self.lbox.add(self.label)

		self.table.attach(self.lbox, 1, 2, 0, 1,
			xoptions=gtk.FILL|gtk.EXPAND,
			yoptions=gtk.FILL|gtk.EXPAND)

		# Create Description label
		self.descr = gtk.Label()
		self.descr.set_properties(
			xalign=0.05,
			yalign=0.2,
			selectable=True,
			use_markup=True,
			width_chars=30,
			wrap=True)

		# Add Description to table
		self.dbox = gtk.EventBox()
		self.dbox.add(self.descr)

		self.table.attach(self.dbox, 1, 2, 1, 2,
			xoptions=gtk.FILL|gtk.EXPAND,
			yoptions=gtk.FILL|gtk.EXPAND)

		def mod_bg(w, c):
			# for debugging purposes
			if False:
				s = w.get_style().copy()
				s.bg[gtk.STATE_NORMAL] = c
				w.set_style(s)

		mod_bg(self.lbox, gtk.gdk.Color("#FF0000"))
		mod_bg(self.dbox, gtk.gdk.Color("#00FF00"))
		mod_bg(self.ibox, gtk.gdk.Color("#0000FF"))

		self.add_with_viewport(self.table)

		if com.sushi.connected:
			self.sushi_connected_cb(com.sushi)
		else:
			self.sushi_disconnected_cb(com.sushi)

		com.sushi.g_connect("maki-connected", self.sushi_connected_cb)
		com.sushi.g_connect("maki-disconnected", self.sushi_disconnected_cb)

	def sushi_connected_cb(self, sushi):
		s = _("You are connected to maki. The next step "
				"is to connect to a server via the server "
				"dialog in the tekka menu.")
		self.descr.set_markup(s)

	def sushi_disconnected_cb(self, sushi):
		s = _("You are not connected to maki. "
				"Without maki you can not connect to "
				"servers or write messages.\n\n"
				"If you are having problems running maki "
				"visit http://sushi.ikkoku.de/ and look whether there is "
				"a solution for your problem. Otherwise, feel free "
				"to ask for support.")
		self.descr.set_markup(s)




"""
Copyright (c) 2009 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

"""
IRC color specifications
"""

import re
import gtk

from gettext import gettext as _

from .. import config
from ..typecheck import types

from . import code
from . import escape

from ..lib import contrast
from .. import gui

COLOR_PATTERN = "([0-9]{1,2})(,[0-9]{1,2}){0,1}.*"
COLOR_TABLE =  {
			 0: contrast.CONTRAST_COLOR_WHITE,
			 1: contrast.CONTRAST_COLOR_BLACK,
			 2: contrast.CONTRAST_COLOR_BLUE,
			 3: contrast.CONTRAST_COLOR_DARK_GREEN,
			 4: contrast.CONTRAST_COLOR_DARK_RED,
			 5: contrast.CONTRAST_COLOR_LIGHT_BROWN,
			 6: contrast.CONTRAST_COLOR_PURPLE,
			 7: contrast.CONTRAST_COLOR_ORANGE,
			 8: contrast.CONTRAST_COLOR_YELLOW,
			 9: contrast.CONTRAST_COLOR_LIGHT_GREEN,
			10: contrast.CONTRAST_COLOR_CYAN,
			11: contrast.CONTRAST_COLOR_AQUA,
			12: contrast.CONTRAST_COLOR_LIGHT_BLUE,
			13: contrast.CONTRAST_COLOR_MAGENTA,
			14: contrast.CONTRAST_COLOR_GREY,
			15: contrast.CONTRAST_COLOR_LIGHT_GREY
		}
COLOR_NAMES = {
			 0: _("white"),
			 1: _("black"),
			 2: _("blue"),
			 3: _("dark green"),
			 4: _("dark red"),
			 5: _("light brown"),
			 6: _("purple"),
			 7: _("orange"),
			 8: _("yellow"),
			 9: _("light green"),
			10: _("cyan"),
			11: _("aqua"),
			12: _("light blue"),
			13: _("magenta"),
			14: _("gray"),
			15: _("light gray")
		}


def _get_output_bg_color():
	return gui.widgets.get_object("output").get_style().base[
		gtk.STATE_NORMAL]

def _get_output_fg_color():
	return gui.widgets.get_object("output").get_style().fg[NORMAL]

def get_widget_base_color(widget):
	return widget.get_style().base[gtk.STATE_NORMAL]


@types (msg = basestring)
def parse_color_codes_to_tags(msg):
	""" Parse the mIRC color format ^Cn[,m] and convert it
		to the intern handled <font></font> tag.
		Convert the numbers n and m into contrast color codes
		and use them as foreground/background.
	"""
	def get_gdk_color(ccolor):
		return contrast.contrast_render_foreground_color(
			_get_output_bg_color(), ccolor)

	last_i = -1
	count = 0 # openend <font>

	# initialize attributes self.pattern / self.color_table
	self = code.init_function_attrs(
		parse_color_codes_to_tags,

		pattern 	= re.compile(chr(3)+COLOR_PATTERN),
		color_table = COLOR_TABLE)

	while True:
		try:
			i = msg.index(chr(3), last_i+1)
		except ValueError:
			break

		match = self.pattern.match(msg[i:i+6])

		if match:
			groups = match.groups()
			tag = "<span"

			if count != 0:
				# close the previous color
				tag = "</span>" + tag
				count -= 1

			try:
				fg = self.color_table[int(groups[0])]
				fg = get_gdk_color(fg)
			except (KeyError, TypeError):
				fg = None
			else:
				tag += " foreground='%s'" % fg

			try:
				bg = self.color_table[int(groups[1][1:])]
				bg = get_gdk_color(bg)
			except (KeyError, TypeError):
				bg = None
			else:
				tag += " background='%s'" % bg

			tag += ">"
			skip_len = 1 + (groups[0] and len(groups[0]) or 0) \
				+ (groups[1] and len(groups[1]) or 0)
			msg = msg[:i] + tag + msg[i+skip_len:]

			count += 1

		else:
			if count > 0:
				# single ^C, if there's an open tag, close it
				msg = msg[:i] + "</span>" + msg[i+1:]
				count -= 1

		last_i = i

	if count != 0:
		# make sure the <font> is closed.
		msg = msg + "</span>"

	return msg


@types (s = basestring)
def parse_color_codes_to_markups(s):
	""" convert color codes to color markups (%C) and escape
		every % in s with %%.
	"""
	s = s.replace("%", "%%")
	return s.replace(chr(3), "%C")


@types (s = basestring)
def parse_color_markups_to_codes(s):
	""" split s for %C markups and parse the numbers following.
		After parsing, return the new string.
	"""
	s_split = escape.unescape_split("%C", s, escape_char="%")
	return chr(3).join(s_split)


@types (text = basestring)
def strip_color_codes(text):
	""" strip all color codes (chr(3)) and the following numbers """

	pattern = re.compile("\003([0-9]{1,2}(,[0-9]{1,2})?)?")

	start = 0

	while True:
		result = pattern.search(text, start)
		if not result:
			break
		text = text[:result.start()] + text[result.end():]
		start = result.end()

	return text

# Text coloring


def is_contrast_color(value):
	""" checks if the given value is a contrast color or a RGB color """
	return isinstance(value, basestring) and value[0] != "#"


def get_color_by_key(key, bgcolor=None):
	""" get the configured color for the given key as GdkColor.
		The key is defined in config section "colors".

		Example: get_color_by_key("last_log") -> gtk.gdk.Color("#dddddd")

		Note that str(gtk.gdk.Color("#000")) == "#000".

		If bgcolor is None, the bgcolor is retrieved
		by _get_output_bg_color.
	"""
	cvalue = config.get("colors",key)

	if not bgcolor:
		bgcolor = _get_output_bg_color()

	if cvalue == None:
		raise KeyError, "Unknown color key: '%s'" % (key)

	if cvalue[0] == "#":
		return gtk.gdk.Color(cvalue)
	elif key == "rules_color" and config.is_default("colors",key):
		return gui.widgets.get_object("output").get_style().base[
					gtk.STATE_INSENSITIVE]
	else:
		return contrast.contrast_render_foreground_color(
			bgcolor, int(cvalue))


def get_nick_color(nick):
	"""
		Returns a static color for the nick given.
		The returned color depends on the color mapping
		set in config module.
	"""
	def pick_nick_color(colors, nick):
		return colors[sum([ord(n) for n in nick]) % len(colors)]

	if not config.get_bool("tekka","color_text"):
		return _get_output_fg_color()

	if not config.get_bool("colors", "nick_contrast_colors"):
		# pick a color out of the user defined list

		colors = config.get_list("colors", "nick_colors", [])
		color = pick_nick_color(colors, nick)

		return color
	else:
		# pick a contrast color

		bg_color = _get_output_bg_color()
		color = pick_nick_color(contrast.colors[:-1], nick)
		r = contrast.contrast_render_foreground_color(bg_color, color)

		return r


def get_text_color(nick):
	"""
		Same as color.get_nick_color but for text and defaults
		to another value (text_message)
	"""
	if not config.get_bool("tekka","color_text"):
		return _get_output_fg_color()

	colors = contrast.colors[:-1]
	if not colors or not config.get_bool("tekka","color_nick_text"):
		return get_color_by_key("text_message")

	bg_color = _get_output_bg_color()
	color = colors[sum([ord(n) for n in nick]) % len(colors)]

	r = contrast.contrast_render_foreground_color(bg_color, color)
	return r


def colorize_message(msgtype, message):
	if not config.get_bool("tekka", "color_text"):
		return message

	else:
		return "<font foreground='%s'>%s</font>" % (
					get_color_by_key("text_%s" % msgtype),
					message)






"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk

from ._builder import widgets

from ._status_manager import status

from . import dialogs



"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

from gtk import main_quit
from gettext import gettext as _

from .. import com
from .. import config
from .. import gui


def ignore_on_welcome(on_welcome_cb=None, on_not_welcome_cb=None):
	def decorator(fun):
		def new_fun(*args, **kwargs):
			if gui.mgmt.is_welcome_screen():
				if on_welcome_cb: on_welcome_cb()
				return
			if on_not_welcome_cb: on_not_welcome_cb()
			return fun(*args, **kwargs)
		return new_fun
	return decorator


def _reset_item(item_name):
	def cb():
		if hasattr(cb, "locked") and cb.locked:
			return
		item = gui.widgets.get_object(item_name)
		cb.locked = True
		item.set_property("active", not item.get_active())
		cb.locked = False
	return cb


class MenuContextType(object):

	def __init__(self, name = "", widgets = None):
		if widgets == None:
			raise ValueError, "%s: widgets is None" % (self, )

		self.widgets = widgets
		self.signals = []
		self.menu = widgets.get_object(name)


	def __getattribute__(self, attr):
		"""
		if (attr[0] != "_"
		and not attr in ("widgets","signals","menu")
		and not attr in self.signals):
				return getattr(self.menu, attr)
		"""
		try:
			return super(MenuContextType,self).__getattribute__(attr)
		except AttributeError:
			return getattr(self.menu, attr)


class MainMenuContext(MenuContextType):


	class TekkaMenuContext(MenuContextType):

		""" tekka menu
			- Connect
			- Quit
		"""

		def __init__(self, *args, **kwargs):
			MenuContextType.__init__(self, *args, **kwargs)

		def connect_activate_cb(self, item):
			""" show up the server dialog or an error message
				if no connection to maki is given """
			def server_dialog_callback(server_list):
				if server_list:
					for server in server_list:
						com.sushi.connect(server)

			try:
				gui.dialogs.show_dialog(
					"servers", server_dialog_callback, need_sushi = True)
			except com.NoSushiError as e:
				gui.mgmt.show_inline_message(
					"NoSushiError", e.args[0], dtype="error")

		def quit_activate_cb(self, item):
			main_quit()


	class MakiMenuContext(MenuContextType):

		""" maki menu
			- Connect
			- Disconnect
			- Shutdown
		"""

		def __init__(self, *args, **kwargs):
			MenuContextType.__init__(self, *args, **kwargs)

		def connect_activate_cb(self, item):
			com.connect()

		def disconnect_activate_cb(self, item):
			com.disconnect()

		def shutdown_activate_cb(self, item):
			if not com.sushi.connected:
				gui.mgmt.show_maki_connection_error(
					_("No connection to maki."),
					_("You can't shutdown maki. You're not connected."))
			else:
				com.sushi.shutdown(config.get(
					"chatting",
					"quit_message", ""))


	class ViewMenuContext(MenuContextType):

		""" View menu
			- Show general output
			- Show side pane
			- Show status bar
			- Show status icon
			- Show topic bar
		"""

		def __init__(self, *args, **kwargs):
			MenuContextType.__init__(self, *args, **kwargs)

		def apply_visibility_settings(self):
			""" read config, apply visibilites on widgets """

			def apply_visibility(wname, cvalue):
				button = self.widgets.get_object(wname)
				if config.get_bool("tekka", cvalue):
					button.set_active(True)
				else:
					button.set_active(False)

			apply_visibility("view_general_output_item",
							 "show_general_output")
			apply_visibility("view_side_pane_item", "show_side_pane")
			apply_visibility("view_status_bar_item", "show_status_bar")
			apply_visibility("view_status_icon_item", "show_status_icon")
			apply_visibility("view_topic_bar_item", "show_topic_bar")

		@ignore_on_welcome(
			on_welcome_cb=_reset_item("view_general_output_item"))
		def showGeneralOutput_toggled_cb(self, item):
			""" toggle visibility of general output """
			gui.mgmt.visibility.show_general_output(item.get_active())
			config.set("tekka","show_general_output",str(item.get_active()))

		@ignore_on_welcome(on_welcome_cb=_reset_item("view_side_pane_item"))
		def showSidePane_toggled_cb(self, item):
			""" toggle visibility of side pane """
			gui.mgmt.visibility.show_side_pane(item.get_active())
			config.set("tekka", "show_side_pane", str(item.get_active()))

		@ignore_on_welcome(
			on_welcome_cb=_reset_item("view_status_bar_item"))
		def showStatusBar_toggled_cb(self, item):
			""" toggle visibility of status bar """
			gui.mgmt.visibility.show_status_bar(item.get_active())
			config.set("tekka", "show_status_bar", str(item.get_active()))

		def showStatusIcon_toggled_cb(self, item):
			""" toggle visibility of status icon """
			gui.mgmt.visibility.show_status_icon(item.get_active())
			config.set("tekka", "show_status_icon", str(item.get_active()))

		@ignore_on_welcome(on_welcome_cb=_reset_item("view_topic_bar_item"))
		def showTopicBar_toggled_cb(self, item):
			""" toggle visibility of topic bar """
			gui.mgmt.visibility.show_topic_bar(item.get_active())
			config.set("tekka", "show_topic_bar", str(item.get_active()))


	class ToolsMenuContext(MenuContextType):

		""" Tools menu
			- Channel List
			- File transfers (DCC)
			- Plugins
			- Debug
			- Preferences
		"""

		def __init__(self, *args, **kwargs):
			MenuContextType.__init__(self, *args, **kwargs)

		def show_no_sushi_error(self, exp):
			gui.mgmt.show_inline_message(
				_("No connection to maki."),
				exp.args[0],
				dtype="error")

		def channelList_activate_cb(self, item):
			""" Show the channel list dialog or display an error message
				if there's no connection to maki or no server is active """
			sTab,cTab = gui.tabs.get_current_tabs()

			if not sTab:
				gui.mgmt.show_inline_message(
					_("tekka could not determine server."),
					_("There is no active server. Click on "
					  "a server tab or a child of a server "
					  "tab to activate the server."),
					dtype="error")

			else:
				try:
					gui.dialogs.show_dialog(
					"channelList", sTab.name, need_sushi = True)
				except com.NoSushiError as e:
					self.show_no_sushi_error(e)

		def dcc_activate_cb(self, item):
			""" show file transfers dialog """
			try:
				gui.dialogs.show_dialog("dcc", need_sushi = True)
			except com.NoSushiError as e:
				self.show_no_sushi_error(e)

		def plugins_activate_cb(self, item):
			gui.dialogs.show_dialog("plugins")

		def debug_activate_cb(self, item):
			gui.dialogs.show_dialog("debug")

		def preferences_activate_cb(self, item):
			gui.dialogs.show_dialog("preferences")


	class HelpMenuContext(MenuContextType):

		""" Help menu
			- IRC color table
			- About
		"""

		def __init__(self, *args, **kwargs):
			MenuContextType.__init__(self, *args, **kwargs)

		def colors_activate_cb(self, item):
			gui.dialogs.show_dialog("colorTable")

		def about_activate_cb(self, item):
			widgets = gui.builder.load_dialog("about")

			d = widgets.get_object("aboutDialog")
			d.connect("response", lambda d,i: d.destroy())
			d.show_all()


	def __init__(self, *args, **kwargs):
		MenuContextType.__init__(self, *args, **kwargs)

		self.tekka = self.TekkaMenuContext(
			name="tekka_menu",
			widgets=self.widgets)

		self.maki = self.MakiMenuContext(
			name="maki_menu",
			widgets=self.widgets)

		self.view = self.ViewMenuContext(
			name="view_menu",
			widgets=self.widgets)

		self.tools = self.ToolsMenuContext(
			name="tools_menu",
			widgets=self.widgets)

		self.help = self.HelpMenuContext(
			name="help_menu",
			widgets=self.widgets)




# coding:UTF-8
"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import gobject
import logging
import string
import time as mtime

from dbus import UInt64
from gettext import gettext as _

from tekka import com
from tekka import config
from tekka import signals
from tekka import gui

from tekka.lib import contrast

from tekka.com import sushi, parse_from
from tekka.signals import connect_signal
from tekka.lib import key_dialog
from tekka.lib import dcc_dialog
from tekka.lib import inline_dialog

from tekka.helper import code
from tekka.helper import color
from tekka.helper import markup
from tekka.typecheck import types

init = False


def setup():
	sushi.g_connect("maki-connected", maki_connected_cb)
	sushi.g_connect("maki-disconnected", maki_disconnected_cb)


def maki_connected_cb(sushi):
	global init

	if init == False:
		# Message-Signals
		connect_signal("message", userMessage_cb)
		connect_signal("notice", userNotice_cb)
		connect_signal("action", userAction_cb)
		connect_signal("away_message", userAwayMessage_cb)
		connect_signal("ctcp", userCTCP_cb)
		connect_signal("error", userError_cb)

		# action signals
		connect_signal("part", userPart_cb)
		connect_signal("join", userJoin_cb)
		connect_signal("names", userNames_cb)
		connect_signal("quit", userQuit_cb)
		connect_signal("kick", userKick_cb)
		connect_signal("nick", userNick_cb)
		connect_signal("user_away", userAway_cb)
		connect_signal("mode", userMode_cb)
		connect_signal("oper", userOper_cb)

		# Server-Signals
		connect_signal("connect", serverConnect_cb)
		connect_signal("connected", serverConnected_cb)
		connect_signal("motd", serverMOTD_cb)
		connect_signal("dcc_send", dcc_send_cb)

		# Channel-Signals
		connect_signal("topic", channelTopic_cb)
		connect_signal("banlist", channelBanlist_cb)

		# Maki signals
		connect_signal("shutdown", makiShutdown_cb)

		init = True

	_add_servers()


def maki_disconnected_cb(sushi):
	pass


@types (server = basestring)
def _setup_server(server):
	tab = gui.tabs.create_server(server)

	gui.tabs.add_tab(None, tab,
		update_shortcuts = config.get_bool("tekka","server_shortcuts"))

	return tab


def _add_servers():
	""" Adds all servers to tekka which are reported by maki. """
	# in case we're reconnecting, clear all stuff
	gui.widgets.get_object("tab_store").clear()

	for server in sushi.servers():
		tab = _setup_server(server)
		tab.connected = True
		_add_channels(tab)

	try:
		toSwitch = gui.tabs.get_all_tabs()[1]
	except IndexError:
		return
	else:
		gui.tabs.switch_to_path(toSwitch.path)


def _add_channels(server_tab):
	"""
		Adds all channels to tekka wich are reported by maki.
	"""
	channels = sushi.channels(server_tab.name)

	for channel in channels:

		add = False
		nicks, prefixes = sushi.channel_nicks(server_tab.name, channel)

		tab = gui.tabs.search_tab(server_tab.name, channel)

		if not tab:
			tab = gui.tabs.create_channel(server_tab, channel)
			add = True

		tab.nickList.clear()
		tab.nickList.add_nicks(nicks, prefixes)

		for nick in nicks:
			# FIXME inefficient → nicks, prefixes, aways = …?
			tab.nickList.set_away(nick, sushi.user_away(server_tab.name, nick))

		tab.topic = sushi.channel_topic(server_tab.name, channel)
		tab.topicsetter = ""

		if tab.is_active():
			gui.set_topic(markup.markup_escape(tab.topic))
			gui.mgmt.set_user_count(
				len(tab.nickList),
				tab.nickList.get_operator_count())

		# TODO: handle topic setter
		tab.joined = True
		tab.connected = True

		if add:
			gui.tabs.add_tab(server_tab, tab, update_shortcuts = False)
			tab.print_last_log()

		topic = sushi.channel_topic(server_tab.name, channel)
		_report_topic(mtime.time(), server_tab.name, channel, topic)

	gui.shortcuts.assign_numeric_tab_shortcuts(gui.tabs.get_all_tabs())

def isHighlighted (server_tab, text):
	def has_highlight(text, needle):
		punctuation = string.punctuation + " \n\t"
		needle = needle.lower()
		ln = len(needle)
		for line in text.split("\n"):
			line = line.lower()
			i = line.find(needle)
			if i >= 0:
				if (line[i-1:i] in punctuation
				and line[ln+i:ln+i+1] in punctuation):
					return True
		return False

	highlightwords = config.get_list("chatting", "highlight_words", [])
	highlightwords.append(server_tab.nick)

	for word in highlightwords:
		if has_highlight(text, word):
			return True
	return False

def action_nick_color(nick):
	""" return the nick color if color_action_nicks is activated,
		otherwise return the default text color
	"""
	if config.get_bool("colors", "color_action_nicks"):
		return color.get_nick_color(nick)
	return color.get_color_by_key("text_action")

@types (server = basestring, name = basestring)
def _createTab (server, name):
	""" check if tab exists, create it if not, return the tab """
	server_tab = gui.tabs.search_tab(server)

	if not server_tab:
		raise Exception("No server tab in _createTab(%s, %s)" % (server,name))

	tab = gui.tabs.search_tab(server, name)

	if not tab:
		if name[0] in server_tab.support_chantypes:
			tab = gui.tabs.create_channel(server_tab, name)
		else:
			tab = gui.tabs.create_query(server_tab, name)

		tab.connected = True
		gui.tabs.add_tab(server_tab, tab)
		tab.print_last_log()

	if tab.name != name:
		# the name of the tab differs from the
		# real nick, correct this.
		tab.name = name

	return tab

def _getPrefix(server, channel, nick):
	tab = gui.tabs.search_tab(server, channel)

	if tab and tab.is_channel():
		return tab.nickList.get_prefix(nick)
	else:
		return ""

@types (tab = gui.tabs.TekkaTab, what = basestring, own = bool)
def _hide_output(tab, what, own = False):
	""" Returns bool.
		Check if the message type determined by "what"
		shall be hidden or not.
		tab should be a TekkaServer, -Channel or -Query
	"""
	if type(tab) == gui.tabs.TekkaChannel:
		cat = "channel_%s_%s" % (
					tab.server.name.lower(),
					tab.name.lower())
	elif type(tab) == gui.tabs.TekkaQuery:
		cat = "query_%s_%s" % (
					tab.server.name.lower(),
					tab.name.lower())
	else:
		return False

	hide = what in config.get_list(cat, "hide", [])
	hideOwn = what in config.get_list(cat, "hide_own", [])

	return ((hide and not own)
		or (own and hideOwn)
		or (hide and own and not hideOwn))

@types (servertab = gui.tabs.TekkaServer, tab = gui.tabs.TekkaTab,
	what = basestring, own = bool)
def _show_output_exclusive(servertab, tab, what, own = False):
	""" Returns bool.
		Determine if the message identified by -what- shall
		be shown in tab -tab- or not.
		-servertab- is not used at the moment.
	"""
	return not _hide_output(tab, what, own = own)

""" Server callbacks """

def serverConnect_cb(time, server):
	"""
		maki is connecting to a server.
	"""
	gui.mgmt.set_useable(True)

	tab = gui.tabs.search_tab(server)

	if not tab:
		tab = _setup_server(server)

	if tab.connected:
		tab.connected = False

		channels = gui.tabs.get_all_tabs(servers = [server])[1:]

		if channels:
			for channelTab in channels:
				if channelTab.is_channel():
					channelTab.joined=False
				channelTab.connected=False

	tab.write(time, "Connecting...", msgtype=gui.tabs.ACTION)

	gui.status.set_visible("connecting", "Connecting to %s" % server)

def serverConnected_cb(time, server):
	"""
		maki connected successfuly to a server.
	"""
	tab = gui.tabs.search_tab(server)

	if not tab:
		tab = _setup_server(server)

	tab.connected = True

	# iterate over tabs, set the connected flag to queries
	for query in [tab for tab in gui.tabs.get_all_tabs(
	servers = [server])[1:] if tab.is_query()]:
		query.connected = True

	tab.write(time, "Connected.", msgtype=gui.tabs.ACTION)

def serverMOTD_cb(time, server, message, first_time = {}):
	""" Server is sending a MOTD.
		Channes are joined 3s after the end of the
		MOTD so at the end of the MOTD, make sure
		that the prefixes and chantypes are read
		correctly.
	"""
	if not first_time.has_key(server):
		tab = gui.tabs.search_tab(server)

		if not tab:
			tab = _setup_server(server)
		else:
			tab.update()

		gui.status.unset("connecting")
		tab.connected = True
		first_time[server] = tab

	if not message:
		# get the prefixes for the server to make
		# sure they are correct
		tab = first_time[server]
		tab.support_prefix = sushi.support_prefix(server)
		tab.support_chantypes = sushi.support_chantypes(server)
		del first_time[server]

	else:
		first_time[server].write(time, markup.escape(message),
			no_general_output = True)

""" Callbacks for channel interaction """

def _report_topic(time, server, channel, topic):
	message = _(u"• Topic for %(channel)s: %(topic)s") % {
		"channel": channel,
		"topic": markup.escape(topic) }

	tab = gui.tabs.search_tab(server, channel)

	if not tab:
		raise Exception, "%s:%s not found." % (server, channel)

	tab.write(time, message, gui.tabs.ACTION, no_general_output=True)

def channelTopic_cb(time, server, from_str, channel, topic):
	"""
		The topic was set on server "server" in channel "channel" by
		user "nick" to "topic".
		Apply this!
	"""
	nick = parse_from(from_str)[0]
	serverTab, channelTab = gui.tabs.search_tabs(server, channel)

	if not channelTab:
		raise Exception("Channel %s does not exist but "
						"emits topic signal." % channel)

	channelTab.topic = topic
	channelTab.topicsetter = nick

	if channelTab == gui.tabs.get_current_tab():
		gui.mgmt.set_topic(markup.markup_escape(topic))

	if not nick:
		# just reporting the topic.
		_report_topic(time, server, channel, topic)

	else:
		if nick == serverTab.nick:
			message = _(u"• You changed the topic to %(topic)s.")
		else:
			message = _(u"• %(nick)s changed the topic to %(topic)s.")

		channelTab.write(time, message % {
								"nick": nick,
								"topic": markup.escape(topic) },
						gui.tabs.ACTION)

def channelBanlist_cb(time, server, channel, mask, who, when):
	"""
		ban list signal.
	"""
	self = code.init_function_attrs(channelBanlist_cb,
		tab = gui.tabs.search_tab(server, channel))

	if not mask and not who and when == -1:
		self.tab.write(time, "End of banlist.", gui.tabs.ACTION)
		code.reset_function_attrs(channelBanlist_cb)

	else:
		timestring = mtime.strftime(
			"%Y-%m-%d %H:%M:%S",
			mtime.localtime(when))

		self.tab.write(
			time,
			"%s by %s on %s" % (
				markup.escape(mask),
				markup.escape(who),
				markup.escape(timestring)),
			gui.tabs.ACTION)

""" Callbacks of maki signals """

def makiShutdown_cb(time):
	gui.mgmt.myPrint("Maki is shut down!")
	gui.mgmt.set_useable(False)

""" Callbacks for users """

def userAwayMessage_cb(timestamp, server, nick, message):
	"""
		The user is away and the server gives us the message he left
		for us to see why he is away and probably when he's back again.
	"""
	tab = gui.tabs.get_current_tab()


	# XXX:  you can still write /msg <nick> and get an away message
	# XXX:: in the query window. This would be a more complex fix.
	try:
		tab.printed_away_message
	except AttributeError:
		print_it = True
	else:
		print_it = not tab.printed_away_message

	if print_it:
		tab.write(
			timestamp,
			_(u"• %(nick)s is away (%(message)s).") % {
				"nick": nick,
				"message": markup.escape(message)},
			gui.tabs.ACTION)

		if tab and tab.name == nick:
			tab.printed_away_message = True


def userMessage_cb(timestamp, server, from_str, channel, message):
	"""
		PRIVMSGs are coming in here.
	"""
	nick = parse_from(from_str)[0]
	(server_tab, channel_tab) = gui.tabs.search_tabs(server, channel)

	if server_tab == None:
		return # happens if the target server does not exist

	if nick.lower() == server_tab.nick.lower():
		ownMessage_cb(timestamp, server, channel, message)
		return

	elif channel.lower() == server_tab.nick.lower():
		userQuery_cb(timestamp, server, from_str, message)
		return

	message = markup.escape(message)

	if isHighlighted(server_tab, message):
		# set mode to highlight and disable setting
		# of text color for the main message (would
		# override channelPrint() highlight color)

		type = gui.tabs.HIGHMESSAGE
		messageString = message
		gui.mgmt.set_urgent(True)
	else:
		# no highlight, normal message type and
		# text color is allowed.

		type = gui.tabs.MESSAGE
		messageString = "<font foreground='%s'>%s</font>" % (
			color.get_text_color(nick), message)

	channel_tab.write(timestamp,
		"&lt;%s<font foreground='%s' weight='bold'>%s</font>&gt; %s" % (
			_getPrefix(server, channel, nick),
			color.get_nick_color(nick),
			markup.escape(nick),
			messageString,
		), type, group_string=nick)


def ownMessage_cb(timestamp, server, channel, message):
	"""
		The maki user wrote something on a channel or a query
	"""
	tab = _createTab(server, channel)
	nick = gui.tabs.search_tab(server).nick

	tab.write(timestamp,
			"&lt;%s<font foreground='%s' weight='bold'>%s</font>&gt;"
			" <font foreground='%s'>%s</font>" % (
				_getPrefix(server, channel, nick),
				color.get_color_by_key("own_nick"),
				nick,
				color.get_color_by_key("own_text"),
				markup.escape(message)), group_string=nick)


def userQuery_cb(timestamp, server, from_str, message):
	"""
		A user writes to us in a query.
	"""
	nick = parse_from(from_str)[0]

	tab = _createTab(server, nick)

	if isHighlighted(tab.server, message):
		mtype = gui.tabs.HIGHMESSAGE
	else:
		mtype = gui.tabs.MESSAGE

	tab.write(timestamp,
		"&lt;<font foreground='%s' weight='bold'>%s</font>&gt; %s" % (
			color.get_nick_color(nick),
			markup.escape(nick),
			markup.escape(message)
		), mtype, group_string=nick)

	# queries are important
	gui.mgmt.set_urgent(True)


def userMode_cb(time, server, from_str, target, mode, param):
	"""
		Mode change on target from nick detected.
		nick and param are optional arguments and
		can be empty.

		As nemo:
			/mode #xesio +o nemo
		will result in:
			userMode(<time>,<server>,"nemo","#xesio","+o","nemo")
	"""

	def n_updatePrefix(tab, nick, mode):
		""" checks if the mode is a prefix-mode (e.g. +o)
			If so, the prefix of the nick `nick` in channel `channel`
			will be updated (fetched).
		"""
		if not nick:
			return

		if mode[1] in tab.server.support_prefix[0]:
			tab.nickList.set_prefix(nick,
				sushi.user_channel_prefix(tab.server.name, tab.name, nick))

			if tab.is_active():
				gui.mgmt.set_user_count(len(tab.nickList),
					tab.nickList.get_operator_count())

	nick = parse_from(from_str)[0]

	# nick: /mode target +mode param
	server_tab = gui.tabs.search_tab(server)

	if not nick:
		# only a mode listing
		server_tab.current_write(
			time,
			_("• Modes for %(target)s: %(mode)s") % {
				"target":target,
				"mode":mode},
			gui.tabs.ACTION)

	else:
		actor = nick
		own = (nick == server_tab.nick)

		if own:
			actor = "You"

		tab = gui.tabs.search_tab(server, target)

		if not tab:
			# no channel/query found

			if param: param = " "+param

			if not _hide_output(server_tab, "mode"):

				actor = "<font foreground='%s'>%s</font>" % (
					action_nick_color(actor), actor)
				target = "<font foreground='%s'>%s</font>" % (
					action_nick_color(target), target)

				server_tab.current_write(
					time,
					"• %(actor)s set %(mode)s%(param)s on %(target)s" % {
						"actor":actor,
						"mode":mode,
						"param":param,
						"target":target},
					gui.tabs.ACTION)
		else:
			# suitable channel/query found, print it there

			n_updatePrefix(tab, param, mode)

			type = gui.tabs.ACTION
			victim = target
			own = (target == server_tab.nick)

			if (param == server_tab.nick) or own:
				type = gui.tabs.HIGHACTION
			elif own:
				victim = "you"

			if param: param = " "+param

			if _show_output_exclusive(server_tab, tab, "mode", own = own):

				actor = "<font foreground='%s'>%s</font>" % (
					color.get_nick_color(actor), actor)
				victim = "<font foreground='%s'>%s</font>" % (
					color.get_nick_color(victim), victim)

				tab.write(time,
					"• %(actor)s set %(mode)s%(param)s on %(victim)s." % {
						"actor": actor,
						"mode": mode,
						"param": param,
						"victim": victim},
					type)


def userOper_cb(time, server):
	"""
		yay, somebody gives the user oper rights.
	"""
	server_tab = gui.tabs.search_tab(server)
	server_tab.current_write(time, "• You got oper access.")


def userCTCP_cb(time, server,  from_str, target, message):
	"""
		A user sends a CTCP request to target.
		I don't know a case in which target is not a channel
		and not queried.
	"""
	nick = parse_from(from_str)[0]
	(server_tab, target_tab) = gui.tabs.search_tabs(server, target)

	if nick.lower() == server_tab.nick.lower():
		# we wrote us
		ownCTCP_cb(time, server, target, message)

	elif target.lower() == server_tab.nick.lower():
		# someone wrote us, put in into a query
		queryCTCP_cb(time, server, from_str, message)

	else:
		# normal ctcp
		headline = _("CTCP from %(nick)s to Channel:") % {
			"nick": markup.escape(nick)}

		target_tab.write(time,
			"<font foreground='#00DD33'>%s</font> %s" %	(
				headline, markup.escape(message)))


def ownCTCP_cb(time, server, target, message):
	"""
		The maki user sends a CTCP request to
		a channel or user (target).
	"""
	server_tab, tab = gui.tabs.search_tabs(server, target)

	if tab:
		# valid query/channel found, print it there

		nickColor = color.get_color_by_key("own_nick")
		textColor = color.get_color_by_key("own_text")

		tab.write(time,
			"&lt;CTCP:<font foreground='%s' weight='bold'>%s</font>&gt; "
			"<font foreground='%s'>%s</font>" % (
				nickColor,
				server_tab.nick,
				textColor,
				markup.escape(message)))

	else:
		server_tab.write(time,
			_("CTCP request from you to %(target)s: %(message)s") % {
				"target": markup.escape(target),
				"message": markup.escape(message)})

def queryCTCP_cb(time, server, from_str, message):
	"""
		A user sends us a CTCP request over a query.

		If no query window is open, send it to the server tab.
	"""
	nick = parse_from(from_str)[0]
	(server_tab, tab) = gui.tabs.search_tabs(server, nick)

	if tab:
		tab.write(time,
				"&lt;CTCP:<font foreground='%s' weight='bold'>%s"
				"</font>&gt; <font foreground='%s'>%s</font>" % (
					color.get_nick_color(nick),
					markup.escape(nick),
					color.get_text_color(nick),
					markup.escape(message)))
	else:
		server_tab.current_write(time,
				"&lt;CTCP:<font foreground='%s' weight='bold'>%s"
				"</font>&gt; <font foreground='%s'>%s</font>" % (
					color.get_nick_color(nick),
					markup.escape(nick),
					color.get_text_color(nick),
					markup.escape(message)))

def ownNotice_cb(time, server, target, message):
	"""
		if query channel with ``target`` exists, print
		the notice there, else print it on the current
		channel of the network which is identified by
		`server`
	"""
	server_tab, tab = gui.tabs.search_tabs(server, target)
	ownNickColor = color.get_color_by_key("own_nick")
	ownNick = server_tab.nick

	if tab:
		tab.write(time,
			"&gt;<font foreground='%s' weight='bold'>%s</font>&lt; "
			"<font foreground='%s'>%s</font>" % (
				color.get_nick_color(target),
				markup.escape(target),
				color.get_text_color(target),
				markup.escape(message)))
	else:
		server_tab.current_write(time,
			"&gt;<font foreground='%s' weight='bold'>%s</font>&lt; "
			"<font foreground='%s'>%s</font>" % (
				color.get_nick_color(target),
				markup.escape(target),
				color.get_text_color(target),
				markup.escape(message)))


def queryNotice_cb(time, server, from_str, message):
	"""
		A user sends a notice directly to the maki user.
	"""
	nick = parse_from(from_str)[0]

	(server_tab, tab) = gui.tabs.search_tabs(server, nick)

	if tab:
		if tab.name != nick:
			# correct notation of tab name
			tab.name = nick

	if tab:
		tab.write(time,
				"-<font foreground='%s' weight='bold'>%s</font>- "
				"<font foreground='%s'>%s</font>" % (
					color.get_nick_color(nick),
					markup.escape(nick),
					color.get_text_color(nick),
					markup.escape(message)))
	else:
		server_tab.current_write(time,
				"-<font foreground='%s' weight='bold'>%s</font>- "
				"<font foreground='%s'>%s</font>" % (
					color.get_nick_color(nick),
					markup.escape(nick),
					color.get_text_color(nick),
					markup.escape(message)))


def userNotice_cb(time, server, from_str, target, message):
	""" An incoming notice """
	nick = parse_from(from_str)[0]
	(server_tab, target_tab) = gui.tabs.search_tabs(server, target)

	if nick.lower() == server_tab.nick.lower():
		# we wrote that notice
		ownNotice_cb(time, server, target, message)
		return

	elif target.lower() == server_tab.nick.lower():
		# it's supposed to be a private (query) message
		queryNotice_cb(time, server, from_str, message)
		return

	message = "-<font foreground='%s' weight='bold'>%s</font>- "\
			  "<font foreground='%s'>%s</font>" % (
				color.get_nick_color(nick),
				markup.escape(nick),
				color.get_text_color(nick),
				markup.escape(message))

	if target_tab == None:
		# global notice
		server_tab.current_write(time, message)
	else:
		# channel/query notice
		target_tab.write(time, message)


def ownAction_cb(time, server, channel, action):

	tab = _createTab(server, channel)
	nick = gui.tabs.search_tab(server).nick

	nickColor = color.get_color_by_key("own_nick")
	textColor = color.get_color_by_key("own_text")

	tab.write(time,
		"<font foreground='%s' weight='bold'>%s</font> "
		"<font foreground='%s'>%s</font>" % (
			nickColor,
			nick,
			textColor,
			markup.escape(action)), group_string=nick)


def actionQuery_cb(time, server, from_str, action):
	""" action in a query """

	nick = parse_from(from_str)[0]

	tab = _createTab(server, nick)

	tab.write(time, "%s %s" % (nick, markup.escape(action)), group_string=nick)

	gui.mgmt.set_urgent(True)


def userAction_cb(time, server, from_str, channel, action):
	""" normal action """

	nick = parse_from(from_str)[0]
	(server_tab, channel_tab) = gui.tabs.search_tabs(server, channel)

	if nick.lower() == server_tab.nick.lower():
		ownAction_cb(time, server, channel, action)
		return

	elif channel.lower() == server_tab.nick.lower():
		actionQuery_cb(time, server, from_str, action)
		return

	action = markup.escape(action)

	if isHighlighted(server_tab, action):
		type = gui.tabs.HIGHACTION
		actionString = action
		gui.mgmt.set_urgent(True)

	else:
		type = gui.tabs.ACTION
		actionString = "<font foreground='%s'>%s</font>" % (
			color.get_text_color(nick), action)

	channel_tab.write(
			time,
			"<font foreground='%s' weight='bold'>%s</font> %s" % (
				color.get_nick_color(nick),
				nick,
				actionString),
			type, group_string=nick)


def userNick_cb(time, server, from_str, newNick):
	"""
	A user (or the maki user) changed it's nick.
	If a query window for this nick on this server
	exists, it's name would be changed.
	"""
	nick = parse_from(from_str)[0]

	# find a query
	server_tab, tab = gui.tabs.search_tabs(server, nick)

	# rename query if found
	if tab and tab.is_query():
		tab.name = newNick

	own = False

	# we changed the nick
	if not nick or nick == server_tab.nick:
		message = _(u"• You are now known as %(newnick)s.")
		server_tab.nick = newNick
		own = True

	# someone else did
	else:
		message = _(u"• %(nick)s is now known as %(newnick)s.")

	# iterate over all channels and look if the nick is
	# present there. If true so rename him in nicklist cache.
	for tab in gui.tabs.get_all_tabs(servers = [server])[1:]:

		if not nick or newNick == server_tab.nick:
			# notification, print everytime
			doPrint = True
		else:
			doPrint = _show_output_exclusive(server_tab, tab, "nick", own)

		if tab.is_channel():
			if (nick in tab.nickList.get_nicks()):
				tab.nickList.modify_nick(nick, newNick)
			else:
				continue

		if tab.is_query() and tab.name != newNick:
			# ignore not associated queries
			continue

		nickString = "<font foreground='%s' weight='bold'>%s</font>" % (
			action_nick_color(nick),
			markup.escape(nick))

		newNickString = "<font foreground='%s' weight='bold'>%s</font>" % (
			action_nick_color(newNick),
			markup.escape(newNick))

		if doPrint:
			tab.write(time,
				message % {
					"nick": nickString,
					"newnick": newNickString
				},
				gui.tabs.ACTION)


def userAway_cb(time, server, from_str, away):
	nick = parse_from(from_str)[0]
	server_tab = gui.tabs.search_tab(server)

	if not server_tab:
		return

	if nick == server_tab.nick:
		if away:
			server_tab.away = "-- Not implemented yet --"
		else:
			server_tab.away = ""

	# iterate over all channels and look if the nick is
	# present there.
	for tab in gui.tabs.get_all_tabs(servers = [server])[1:]:

		if tab.is_channel():
			if nick in tab.nickList.get_nicks():
				tab.nickList.set_away(nick, away)

		elif tab.is_query():
			if tab.name == nick:
				# FIXME
				continue


def userKick_cb(time, server, from_str, channel, who, reason):
	"""
		signal emitted if a user got kicked.
		If the kicked user is ourself mark the channel as
		joined=False
	"""
	nick = parse_from(from_str)[0]
	server_tab, tab = gui.tabs.search_tabs(server, channel)

	if not tab:
		logging.debug("userKick: channel '%s' does not exist." % (channel))
		return

	channelString = markup.escape(channel)

	nickString = "<font foreground='%s' weight='bold'>%s</font>" % (
		action_nick_color(nick), markup.escape(nick))

	reasonString = markup.escape(reason)

	if who == server_tab.nick:
		tab.joined = False

		if _show_output_exclusive(server_tab, tab, "kick", own = True):

			message = _(u"« You have been kicked from %(channel)s "
				u"by %(nick)s (%(reason)s)." % {
					"channel": channelString,
					"nick": nickString,
					"reason": reasonString })

			tab.write(time, message, gui.tabs.HIGHACTION)

	else:
		tab.nickList.remove_nick(who)

		if tab.is_active():
			gui.mgmt.set_user_count(
				len(tab.nickList),
				tab.nickList.get_operator_count())

		if _show_output_exclusive(server_tab, tab, "kick"):

			whoString = "<font foreground='%s' weight='bold'>%s</font>" % (
				color.get_nick_color(who), markup.escape(who))

			message = _(u"« %(who)s was kicked from %(channel)s by "
						u"%(nick)s (%(reason)s).") % {
				"who": whoString,
				"channel": channelString,
				"nick": nickString,
				"reason": reasonString }

			tab.write(time, message, gui.tabs.ACTION)

def userQuit_cb(time, server, from_str, reason):
	"""
	The user identified by nick quit on the server "server" with
	the reason "reason". "reason" can be empty ("").
	If we are the user all channels were set to joined=False and
	the server's connected-flag is set to False (as well as the
	connect-flags of the childs).

	If another user quits on all channels on which the user was on
	a message is generated.
	"""
	server_tab = gui.tabs.search_tab(server)
	nick = parse_from(from_str)[0]

	if not server_tab:
		# tab was closed before
		return

	if nick == server_tab.nick:
		# set the connected flag to False for the server

		server_tab.connected = False

		hideServerPrint = _hide_output(server_tab, "quit", own = True)

		# walk through all channels and set joined = False on them
		channels = gui.tabs.get_all_tabs(servers = [server])[1:]

		if reason:
			message = _(u"« You have quit (%(reason)s).")
		else:
			message = _(u"« You have quit.")

		# deactivate channels/queries
		for channelTab in channels:

			hideChannelPrint = _hide_output(channelTab, "quit",
				own = True)

			if channelTab.is_channel():
				channelTab.joined = False

			channelTab.connected = False

			if not (hideServerPrint or hideChannelPrint):
				channelTab.write(time, message % {
									"reason": reason},
								gui.tabs.ACTION)

	else: # another user quit the network

		hideServerPrint = _hide_output(server_tab, "quit")

		if reason:
			message = _(u"« %(nick)s has quit (%(reason)s).")
		else:
			message = _(u"« %(nick)s has quit.")

		nickString = "<font foreground='%s' weight='bold'>"\
			"%s</font>" % (
				action_nick_color(nick),
				markup.escape(nick))

		reasonString = markup.escape(reason)

		message = message % {
			"nick": nickString,
			"reason": reasonString}

		channels = gui.tabs.get_all_tabs(servers = [server])[1:]

		if not channels:
			logging.debug("No channels but quit reported.. Hum wtf? o.0")
			return

		# print in all channels where nick joined a message
		for channelTab in channels:

			hideChannelPrint = _hide_output(channelTab, "quit")

			if channelTab.is_query():
				# on query with `nick` only print quitmessage

				if (not (hideChannelPrint or hideServerPrint)
				and channelTab.name.lower() == nick.lower()):
					channelTab.write(time, message, gui.tabs.ACTION)

				# skip nickList modification for queries
				continue

			# search for the nick in the channel
			# and print the quit message if the
			# nick was found.
			nickList = channelTab.nickList
			nicks = nickList.get_nicks() or []

			if nick in nicks:
				nickList.remove_nick(nick)

				if channelTab.is_active():
					# update gui display for usercount
					gui.mgmt.set_user_count(len(nickList),
						nickList.get_operator_count())

				if not (hideServerPrint or hideChannelPrint):
					channelTab.write(time, message, gui.tabs.ACTION)


def userJoin_cb(timestamp, server, from_str, channel):
	"""
	A user identified by "nick" joins the channel "channel" on
	server "server.

	If the nick is our we add the channeltab and set properties
	on it, else we generate messages and stuff.
	"""
	nick = parse_from(from_str)[0]
	stab, tab = gui.tabs.search_tabs(server, channel)
	doPrint = False

	if nick == stab.nick:
		# we joined a channel, fetch nicks and topic, create
		# channel and print the log

		if not tab:
			tab = gui.tabs.create_channel(stab, channel)

			if not gui.tabs.add_tab(stab, tab):
				raise Exception, \
					"userJoin_cb: adding tab for channel '%s' failed." % (
					channel)

			tab.print_last_log()

		tab.nickList.clear()

		if tab.is_active():
			gui.mgmt.set_user_count(
				len(tab.nickList),
				tab.nickList.get_operator_count())

		tab.joined = True
		tab.connected = True

		if config.get_bool("tekka","switch_to_channel_after_join"):
			gui.tabs.switch_to_path(tab.path)

		doPrint = _show_output_exclusive(stab, tab, "join", own = True)

		if doPrint:

			nickString = "You"
			message = _(u"» You have joined %(channel)s.")

	else: # another one joined the channel

		if not tab:
			raise Exception, \
				"No tab for channel '%s' in userJoin (not me)."

		doPrint = _show_output_exclusive(stab, tab, "join", own = False)

		if doPrint:
			message = _(u"» %(nick)s has joined %(channel)s.")
			nickString = "<font foreground='%s' weight='bold'>"\
				"%s</font>" % (
					action_nick_color(nick),
					markup.escape(nick))

		tab.nickList.append_nick(nick)

		if tab.is_active():
			gui.mgmt.set_user_count(len(tab.nickList),
				tab.nickList.get_operator_count())

	if doPrint:
		message = message % {
			"nick": nickString,
			"channel": markup.escape(channel) }

		tab.write(timestamp, message, gui.tabs.ACTION)

def userNames_cb(timestamp, server, channel, nicks, prefixes):
	"""
	this signal is called for each nick in the channel.
	remove the nick to make sure it isn't there (hac--workaround),
	add the nick, fetch the prefix for it and at least
	update the user count.
	"""
	server_tab, tab = gui.tabs.search_tabs(server, channel)

	if not server_tab or not tab: # /names for unexisting channel?
		return

	if not nicks:
		# end of list
		tab.nickList.sort_nicks()

	else:
		for i in xrange(len(nicks)):
			# FIXME
			tab.nickList.remove_nick(nicks[i])
			tab.nickList.append_nick(nicks[i], sort=False)

			if prefixes[i]:
				tab.nickList.set_prefix(nicks[i], prefixes[i], sort=False)

			# FIXME inefficient
			tab.nickList.set_away(nicks[i], sushi.user_away(server_tab.name, nicks[i]))

		if tab.is_active():
			gui.mgmt.set_user_count(
				len(tab.nickList),
				tab.nickList.get_operator_count())


def userPart_cb(timestamp, server, from_str, channel, reason):
	"""
	A user parted the channel.

	If we are the user who parted, mark the channel
	as parted (joined=False)
	"""
	nick = parse_from(from_str)[0]

	stab, tab = gui.tabs.search_tabs(server, channel)

	if not tab:
		# tab was closed
		return

	channelString = markup.escape(channel)
	reasonString = markup.escape(reason)

	if nick == stab.nick:
		# we parted

		tab.joined = False

		if _show_output_exclusive(stab, tab, "part", own = True):

			if reason:
				message = _(u"« You have left %(channel)s (%(reason)s).")
			else:
				message = _(u"« You have left %(channel)s.")

			tab.write(timestamp,
						message % {
							"channel": channelString,
							"reason": reasonString },
						gui.tabs.ACTION)

	else: # another user parted

		tab.nickList.remove_nick(nick)

		if tab.is_active():
			gui.mgmt.set_user_count(
				len(tab.nickList),
				tab.nickList.get_operator_count())

		if _show_output_exclusive(stab, tab, "part", False):

			nickString = "<font foreground='%s' weight='bold'>"\
				"%s</font>" % (
				color.get_nick_color(nick), markup.escape(nick))

			if reason:
				message = _(u"« %(nick)s has left %(channel)s "\
					"(%(reason)s).")
			else:
				message = _(u"« %(nick)s has left %(channel)s.")

			tab.write(timestamp,
						message % {
							"nick": nickString,
							"channel": channelString,
							"reason": reasonString},
						gui.tabs.ACTION)


def userError_cb(time, server, domain, reason, arguments):
	if domain == "no_such":
		noSuch(time, server, arguments[0], reason)

	elif domain == "cannot_join":
		cannotJoin(time, server, arguments[0], reason)

	elif domain == "privilege":
		if reason == "channel_operator":
			tab = gui.tabs.search_tab(server, arguments[0])

			message = _(u"• You are not a channel operator.")
			tab.write(time, message)


def noSuch(time, server, target, type):
	""" Signal is emitted if maki can't find the target on the server. """

	(server_tab, tab) = gui.tabs.search_tabs(server, target)

	if type == "nick":
		error = _(u"• %(target)s: No such nick/channel.") % {
			"target": markup.escape(target) }
	elif type == "server":
		error = _(u"• %(target)s: No such server.") % {
			"target": markup.escape(target) }
	elif type == "channel":
		error = _(u"• %(target)s: No such channel.") % {
			"target": markup.escape(target) }

	if tab:
		tab.write(time, error)
	else:
		server_tab.write(time, error)

def cannotJoin(time, server, channel, reason):
	""" The channel could not be joined.
		reason : { l (full), i (invite only), b (banned), k (key) }
	"""
	message = _("Unknown reason")

	if reason == "full":
		message = _("The channel is full.")
	elif reason == "invite":
		message = _("The channel is invite-only.")
	elif reason == "banned":
		message = _("You are banned.")
	elif reason == "key":
		if config.get_bool("tekka", "ask_for_key_on_cannotjoin"):

			def key_dialog_response_cb(dialog, id):
				if id == gtk.RESPONSE_OK:
					sushi.join(server, channel, dialog.entry.get_text())
				dialog.destroy()

			# open a input dialog which asks for the key
			d = key_dialog.KeyDialog(server, channel)
			d.connect("response", key_dialog_response_cb)
			gui.mgmt.show_inline_dialog(d)
			return

		else:
			message = _("You need the correct channel key.")

	server_tab = gui.tabs.search_tab(server)

	server_tab.current_write(time,
		_("You can not join %(channel)s: %(reason)s" % {
			"channel":channel,
			"reason":message
			}))


def dcc_send_cb(time, id, server, sender, filename,
			 size, progress, speed, status):
	"""
	status:
	- 1 << 0 = incoming
	- 1 << 1 = resumed
	- 1 << 2 = running
	- 1 << 3 = error

	"" in (server, sender, filename)
		and 0 in (size, progress, speed, status):
	send was removed
	"""

	def dcc_dialog_response_cb(dialog, id, tid):
		if id == gtk.RESPONSE_OK:
			sushi.dcc_send_accept(tid)
		elif id == gtk.RESPONSE_CANCEL:
			sushi.dcc_send_remove(tid)
		dialog.destroy()


	# create dcc_news new-transfer-cache
	self = code.init_function_attrs(dcc_send_cb,
		dcc_news={},
		dcc_notifies={})

	if (server == "" and sender == "" and filename == ""
	and size == 0 and progress == 0 and speed == 0 and status == 0):

		# send was removed
		logging.debug("filetransfer %d removed." % (id))

		try:
			del self.dcc_news[id]
			del self.dcc_notifies[id]
		except KeyError:
			pass

		return

	logging.debug("status is %d." % (status))

	# import dcc transfer states
	from tekka.helper.dcc import s_new, s_incoming, s_resumable, s_running

	# handle incoming transfer
	if status & s_incoming == s_incoming:

		if status & s_new == s_new:
			# attempt made

			d = dcc_dialog.DCCDialog(
				id, parse_from(sender)[0],
				filename, size,
				resumable = (status & s_resumable == s_resumable))

			self.dcc_news[id] = True

			d.connect("response", dcc_dialog_response_cb, id)
			gui.mgmt.show_inline_dialog(d)

		elif status & s_running and status & s_incoming:
			if not self.dcc_news.has_key(id) and not self.dcc_notifies.has_key(id):
				# notify about auto accepted file transfer
				gui.mgmt.show_inline_message(
					_("Auto accepted file transfer"),
					_("maki auto accepted the following file transfer:\n"
					  "Filename: %(filename)s\n"
					  "Sender: %(sender)s\n"
					  "Size: %(size)s\n"
					  "Server: %(server)s" % {
						  "filename":filename,
						  "sender":parse_from(sender)[0],
						  "size":size,
						  "server":server}),
					dtype="info")

				self.dcc_notifies[id] = True



# coding: UTF-8
"""
Copyright (c) 2009 Michael Kuhn
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import sushi

# tekka-specific
import tekka.config as config
import tekka.gui as gui

import gobject
import gtk
import pynotify
import string
import json

# FIXME configurable highlight words

plugin_info = (
	"Notifies on highlight.",
	"1.1",
	"Michael Kuhn"
)

plugin_options = (
	("targets", "Targets (e.g. Freenode:#sushi-irc)", sushi.TYPE_LIST, ""),
)


class notify (sushi.Plugin):

	def __init__ (self):
		sushi.Plugin.__init__(self, "notify")

		pynotify.init("tekka")

		self.caps = pynotify.get_server_caps()

		try:
			self.pixbuf = gtk.icon_theme_get_default().load_icon("tekka",64,0)
		except:
			self.pixbuf = None

		# FIXME
		self.connect_signal("message", self.message_cb)
		self.connect_signal("action", self.action_cb)

	def unload (self):
		self.disconnect_signal("message", self.message_cb)
		self.disconnect_signal("action", self.action_cb)

	def notify (self, subject, body):
		if gui.mgmt.has_focus():
			return

		notification = pynotify.Notification(subject, body)

		if self.pixbuf:
			notification.set_icon_from_pixbuf(self.pixbuf)

		if "append" in self.caps:
			notification.set_hint_string("append", "allowed")

		if "x-canonical-append" in self.caps:
			notification.set_hint_string("x-canonical-append", "allowed")

		notification.show()

	def escape (self, message):
		# Bold
		message = message.replace(chr(2), "")
		# Underline
		message = message.replace(chr(31), "")

		message = gobject.markup_escape_text(message)

		return message

	def has_highlight(self, text, needle):
		punctuation = string.punctuation + " \n\t"
		needle = needle.lower()
		ln = len(needle)
		for line in text.split("\n"):
			line = line.lower()
			i = line.find(needle)
			if i >= 0:
				if (line[i-1:i] in punctuation
				and line[ln+i:ln+i+1] in punctuation):
					return True
		return False

	def build_tab_name(self, server, target):
		return "%s:%s" % (server, target)

	def notify_target(self, server, target):
		""" return True if the user wants to be notified about text in
			server/target.
		"""
		targets = self.get_config("targets")
		if not targets:
			return []
		return self.build_tab_name(server,target) in json.loads(targets)

	def message_cb (self, timestamp, server, from_str, target, message):
		nick = from_str.split("!")[0]
		own_nick = self.get_nick(server)

		if own_nick:
			own_nick = own_nick.lower()

		if not own_nick:
			return
		elif own_nick == nick.lower():
			return

		def in_notify():
			self.notify(target, "&lt;%s&gt; %s" % (
				nick,
				self.escape(message)))

		if own_nick == target.lower():
			self.notify(nick, self.escape(message))
		elif self.has_highlight(message, own_nick):
			in_notify()
		elif self.notify_target(server, target):
			self.notify("%s:%s:%s" % (server, target, nick),
						self.escape(message))
		else:
			for word in config.get_list("chatting","highlight_words",[]):
				if self.has_highlight(message, word):
					in_notify()
					break

	def action_cb (self, time, server, from_str, target, action):
		nick = from_str.split("!")[0]
		own_nick = self.get_nick(server)

		if own_nick:
			own_nick = own_nick.lower()

		if not own_nick:
			return
		elif own_nick == nick.lower():
			return

		def in_notify():
			self.notify(target, "%s %s" % (nick, self.escape(action)))

		if own_nick == target.lower():
			self.notify(nick, self.escape(action))
		elif self.has_highlight(action, own_nick):
			in_notify()
		elif self.notify_target(server, target):
			self.notify("%s:%s:%s" % (server, target, nick),
						self.escape(action))
		else:
			for word in config.get_list("chatting","highlight_words",[]):
				if self.has_highlight(action, word):
					in_notify()
					break



# coding: UTF-8
"""
Copyright (c) 2009 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import sys
from typecheck import types
from dbus import String, UInt64

import urwid
from urwid import MetaSignals, SimpleListWalker, Text

import connection
import config

def tree_to_list(input_parents, target = None):

	l = []

	for parent in input_parents:

		if target and parent != target:
			continue

		l.append(parent)

		for child in parent.children:
			l.append(child)

	return l

def get_server(tab):
	if not tab:
		return None
	if type(tab) == Server:
		return tab
	else:
		parent = tab.parent
		if not parent:
			return None
		return tab.parent
	return None

class Tab(object):

	__metaclass__ = MetaSignals
	signals = ["child_added", "child_removed", "remove", "connected"]

	_valid_stati = [
		"informative",
		"actions", "actions_highlight","actions_own",
		"messages","messages_highlight","messages_own"]

	def __repr__(self):
		return "<tab: %s:%s:%s>" % (
			type(self).__name__, self._parent, self.name)

	def __str__(self):
		return self.name

	# Properties

	@types(switch = bool)
	def set_connected(self, switch):
		self._connected = switch

		for child in self.children:
			child.set_connected(switch)

		urwid.emit_signal(self, "connected", switch)

	connected = property(lambda x: x._connected, set_connected)

	def set_input_text(self, text):
		self._input_text = text

	input_text = property(lambda x: x._input_text, set_input_text)

	# Methods

	@types(name = (String, str, unicode))
	def __init__(self, name):
		self.name = name

		self._connected = False
		self._parent = None

		self.status = {}
		self.children = []
		self.input_history = None
		self.output_walker = SimpleListWalker([])
		self.input_text = ""
		self.auto_scroll = True

		self.read_line_index = -1

	def set_readline(self):
		if self.read_line_index != -1:
			try:
				del self.output_walker[self.read_line_index]
			except IndexError:
				pass

		line = Text("-"*30)
		line.set_align_mode ("center")
		self.output_walker.append(line)
		self.read_line_index = len(self.output_walker)-1

	def child_added(self, child):
		urwid.emit_signal(self, "child_added", self, child)
		child.set_connected(self.connected)
		self.children.append(child)

	def child_removed(self, child):
		urwid.emit_signal(self, "child_removed", self, child)
		try:
			i = self.children.index(child)
		except ValueError:
			pass
		else:
			del self.children[i]

	def set_parent(self, parent):
		if parent in self.children:
			print "Loop detected in '%s'.set_parent(%s)" % (
				self, parent)
			return

		try:
			self._parent.child_removed(self)
		except AttributeError:
			pass

		self._parent = parent

		try:
			self._parent.child_added(self)
		except AttributeError:
			pass

	parent = property(lambda x: x._parent, set_parent)

	def remove(self):
		""" emit remove signals """
		for child in self.children:
			child.remove()
			self.child_removed(child)
		urwid.emit_signal(self, "remove")
		self.set_parent(None)

	@types(name = str)
	def add_status(self, name):
		if name in self._valid_stati:
			self.status[name] = True

	@types(status = str)
	def has_status(self, status):
		return self.status.has_key(status)

	@types(name = str)
	def remove_status(self, name):
		try:
			del self.status[name]
		except KeyError:
			pass

	def reset_status(self):
		self.status = {}

	def print_last_log(self, lines=0):
		"""	Fetch the given amount of lines of history for
			the channel on the given server and print it to the
			channel's textview.
		"""
		lines = UInt64(lines or config.get(
			"chatting",
			"last_log_lines",
			"0"))

		if type(self) == Server:
			server = self.name
			channel = ""
		else:
			server = self.parent.name
			channel = self.name

		for line in connection.sushi.log(server, channel, lines):
			self.output_walker.append(Text(unicode(line)))

class Server(Tab):

	def __init__(self, name):
		Tab.__init__(self, name)

		self._nick = None
		self._away = ""

		self.support_chantypes = ()
		self.support_prefix = ()

	def update(self):
		self.support_prefix = connection.sushi.support_prefix(self.name)
		self.support_chantypes = connection.sushi.support_chantypes(self.name)

	@types(nick = (str,String))
	def set_nick(self, nick):
		self._nick = nick

	def get_nick(self):
		return self._nick

	@types(msg = str)
	def set_away(self, msg):
		self._away = msg

class NickList(dict):

	def __init__(self):
		dict.__init__(self)

	@types(host = (str, unicode, String))
	def add_nick(self, host):
		(nick,ident,host) = connection.parse_from(host)
		self[unicode(nick)] = (ident,host)

	@types(nick = (str, unicode, String))
	def get_hostmask(self, nick):
		return self[unicode(nick)]

	@types(old = (str, unicode, String),
		new = (str, unicode, String))
	def rename_nick(self, old, new):
		u_old = unicode(old)
		self[unicode(new)] = self[u_old]
		del self[u_old]

	@types(nick = (str, unicode, String))
	def remove_nick(self, nick):
		del self[unicode(nick)]

	@types(nick = (str, unicode, String))
	def has_nick(self, nick):
		return self.has_key(unicode(nick))

class Channel(Tab):

	@types(name = (String,str))
	def __init__(self, name, parent = None):
		Tab.__init__(self, name)
		self.joined = False
		self.set_parent(parent)
		self._topic = ""
		self.nicklist = NickList()

	@types(switch = bool)
	def set_joined(self, switch):
		self.joined = switch

	@types(topic = (String, unicode, str))
	def set_topic(self, topic):
		self._topic = unicode(topic)

	def get_topic(self):
		return self._topic

class Query(Tab):

	def __init__(self, name, parent = None):
		Tab.__init__(self, name)
		self.set_parent(parent)



"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk

from . import widgets
from .. import config
from ..helper import shortcuts

"""
Holds the information which shortcutname has which handler associated
and how it is triggered (the shortcut itself).

At this time everything about shortcuts happens in here but for
future extension possibilities (remapping shortcuts etc.) this
API is provided.

Also it is possible for tekka specific plugin developers to add
shortcuts easily.
"""

_handlers = {
		"clear_outputs": [],
		"output_page_up": [],
		"output_page_down": [],
		"input_clear_line": [],
		"input_search": [],
		"input_search_further": [],
		"input_copy": [],
		"change_topic": [],
		"servertree_previous": [],
		"servertree_next": [],
		"servertree_close": [],
		"show_sidepane": []
	}

def add_handlers(d):
	global _handlers

	for (key, val) in d.items():
		if _handlers.has_key(key):
			_handlers[key].append(val)

def get_shortcut_handler(shortcut):
	global _handlers

	try:
		return _handlers[shortcut]
	except KeyError:
		return None

def associate_handler(short_name, shortcut, widget):
	global _handlers

	try:
		for handler in _handlers[short_name]:
			shortcuts.addShortcut(
				widgets.get_object("main_accel_group"),
				widgets.get_object(widget),
				shortcut,
				handler)

	except KeyError:
		pass

def setup_shortcuts():
	"""
		Set shortcuts to widgets.

		- ctrl + page_up -> scroll to prev tab in server tree
		- ctrl + page_down -> scroll to next tab in server tree
		- ctrl + w -> close the current tab
		- ctrl + l -> clear the output buffer
		- ctrl + u -> clear the input entry
		- ctrl + s -> hide/show the side pane
	"""

	associate_handler("clear_outputs", "<ctrl>l", "input_entry")

	associate_handler("input_clear_line", "<ctrl>u", "input_entry")
	associate_handler("input_search", "<ctrl>f", "input_entry")
	associate_handler("input_search_further", "<ctrl>g", "input_entry")
	associate_handler("input_copy", "<ctrl>c", "input_entry")

	associate_handler("change_topic", "<ctrl>t", "input_entry")

	associate_handler("servertree_previous", "<ctrl>Page_Up", "tabs_view")
	associate_handler("servertree_next", "<ctrl>Page_Down", "tabs_view")
	associate_handler("servertree_close", "<ctrl>w", "tabs_view")

	associate_handler("output_page_up", "Page_Up", "input_entry")
	associate_handler("output_page_down", "Page_Down", "input_entry")

	associate_handler("show_sidepane", "<ctrl>s", "view_side_pane_item")


def assign_numeric_tab_shortcuts(tabList):
	""" assign numeric shortcuts (alt+N) for each
		tab in the list tabs.
	"""

	st = widgets.get_object("tabs_view")
	ag = widgets.get_object("main_accel_group")

	for i in range(1, 10):
		shortcuts.removeShortcut(ag, st, "<alt>%d" % (i))

	c = 1
	for tab in tabList:
		if c == 10:
			break

		if (tab.is_server()
		and not config.get("tekka", "server_shortcuts")):
			continue

		shortcuts.addShortcut(ag, st, "<alt>%d" % (c),
			lambda w, s, p: p.switch_to(), tab)

		c+=1





# coding: UTF-8
"""
Copyright (c) 2009 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import gobject
from threading import Timer # for smooth scrolling

from .htmlbuffer import HTMLBuffer
from ..helper import URLHandler
from .. import config

class OutputTextView(gtk.TextView):


	def __init__(self):
		gtk.TextView.__init__(self,
			HTMLBuffer(handler=URLHandler.URLHandler))

		self.set_property("editable", False)
		self.set_property("can-focus", False)
		self.set_property("wrap-mode", gtk.WRAP_WORD_CHAR)
		self.set_property("cursor-visible", False)

                spacing = int(config.get("tekka", "line_spacing", default=0))
                self.set_line_spacing(spacing)

		self.read_line = ()

		##
		self.smooth_id = None
		#self.smooth_scroll_timer is set in smooth_scroll_to_end
		##

        def set_line_spacing(self, spacing):
            self.set_property("pixels-above-lines", spacing/2)
            self.set_property("pixels-below-lines", spacing/2)


	""" < """
	# Smooth scrolling inspired by Gajim Code

	# smooth scroll constants
	MAX_SCROLL_TIME = 0.4 # seconds
	SCROLL_DELAY = 33 # milliseconds

	"""
	TODO:  optimize the whole code for manual smooth
	TODO:: scrolling even if the slider is set directly
	TODO:: to a position. This needs a replacement for
	TODO:: the current end-mark (the bottom of the buffer)
	"""

	def smooth_scroll(self):
		""" idle add handler for smooth scrolling.
			Returns True if it's going to be recalled.
			Scrolls 1/3rd of the distance to the bottom.

			TODO:  add direction parameter for use with
			TODO:: manual scrolling.
		"""

		parent = self.get_parent()

		if not parent:
			return False

		vadj = parent.get_vadjustment()
		max_val = vadj.upper - vadj.page_size
		cur_val = vadj.get_value()

		# scroll by 1/3rd of remaining distance
		onethird = cur_val + ((max_val - cur_val) / 3.0)

		vadj.set_value(onethird)

		if max_val - onethird < 0.1:
			self._do_smooth_scroll_timeout()
			return False

		return True

	def _smooth_scroll_timeout(self):
		gobject.idle_add(self._do_smooth_scroll_timeout)

	def _do_smooth_scroll_timeout(self):
		""" Timout handler.
			Time's up, if we were done, ok, if not
			and it's remaing space given,
			remove the timer and jump fast forward.
		"""
		if not self.smooth_id:
			# we finished scrolling
			return False

		parent = self.get_parent()

		if parent:
			vadj = parent.get_vadjustment()
			vadj.set_value(vadj.upper - vadj.page_size)

		gobject.source_remove(self.smooth_id)
		self.smooth_id = None

		return False

	def _smooth_scroll_to_end(self):
		""" Call SCROLL_DELAY seconds smooth_scroll
			and scroll 1/3 of the distance.
			If MAX_SCROLL_TIME is reached and we're
			not at the end, fast forward.
		"""
		if None != self.smooth_id:
			# already scrolling
			return False

		self.smooth_id = gobject.timeout_add(self.SCROLL_DELAY,
				self.smooth_scroll)
		self.smooth_scroll_timer = Timer(self.MAX_SCROLL_TIME,
				self._smooth_scroll_timeout)

		self.smooth_scroll_timer.start()

		return False

	def _scroll_to_end(self):
		""" Scroll normally to the end of the buffer """
		# TODO: could be replaced by parent -> vscrollbar -> adj.value = 0
		parent = self.get_parent()
		buffer = self.get_buffer()
		end_mark = buffer.create_mark("end", buffer.get_end_iter(), False)

		self.scroll_to_mark(end_mark, 0, True, 0, 1)

		# reset horizontal scrollbar (do avoid side effects)
		# FIXME:  maybe straight left is not that good for
		# FIXME:: non-western encodings
		if parent:
			adjustment = parent.get_hadjustment()
			adjustment.set_value(0)

		# avoid recalling through idle_add
		return False

	""" > """

	def is_smooth_scrolling(self):
		""" return if we're in the middle of smooth scrolling process """
		return self.smooth_id != None

   	def stop_scrolling(self):
		""" interrupts smooth scrolling procedure """
		if self.smooth_id:
			gobject.source_remove(self.smooth_id)
			self.smooth_id = None
			self.smooth_scroll_timer.cancel()

	def scroll_to_bottom(self, no_smooth = False):
		""" scroll to the end of the textbuffer """
		if config.get_bool("tekka","smooth_scrolling") and not no_smooth:
			self._smooth_scroll_to_end()
		else:
			self._scroll_to_end()

	def get_last_line(self):
		""" returns the last readable line
			(without read_line)
		"""
		buffer = self.get_buffer()
		count = buffer.get_line_count()

		if self.read_line:
			count -= 2
			lineEnd = buffer.get_iter_at_mark(self.read_line[1])
		else:
			lineEnd = buffer.get_end_iter()

		if count <= 0:
			return ""

		lineStart = buffer.get_iter_at_line(count)

		return buffer.get_text(lineStart, lineEnd)

	def set_read_line(self):
		buffer = self.get_buffer()

		if self.read_line:
			markA, markB = self.read_line[1:]
			iterA = buffer.get_iter_at_mark(markA)
			iterB = buffer.get_iter_at_mark(markB)

			if None in (iterA, iterB):
				raise ValueError, "set_read_line: %s,%s in None." % (
					iterA, iterB)

			buffer.delete(iterA, iterB)
			buffer.remove_tag(self.read_line[0], iterA, iterB)

		tag = buffer.create_tag(None,
			justification = gtk.JUSTIFY_CENTER,
			strikethrough = True)

		end_iter = buffer.get_end_iter()
		start_mark = buffer.create_mark(None, end_iter, True)

		buffer.insert_with_tags(end_iter,
			"\n"+" "*int(config.get("tekka","divider_length")), tag)

		end_mark = buffer.create_mark(None, buffer.get_end_iter(), True)

		self.read_line = (tag, start_mark, end_mark)





import gtk

import tekka.main


tekka.main.setup()

from tekka.lib.contrast_color_table import ContrastColorTable

w = gtk.Window()
x = ContrastColorTable()

w.add(x)
w.connect("destroy", lambda *_: gtk.main_quit())
w.show_all()

gtk.main()



import gtk
import pango

from gettext import gettext as _

from ..helper import color
from ..lib import contrast


try:
	from sexy import SpellEntry as _SpellEntry
except ImportError:
	from gtk import Entry as _SpellEntry

class SpellEntry(_SpellEntry):

	__gtype_name__ = "SpellEntry"

	def __init__(self, *args,  **kwargs):
		super(SpellEntry, self).__init__(*args, **kwargs)

		self._ranges = {}

		self.connect("populate-popup",  self._determine_popup_menu)

		self.connect("activate", lambda s,*u: s._apply_color_tags())


	def _determine_popup_menu(self,  widget,  menu,  *user):
		self._popup_menu_handler(menu)


	# TODO:
	# nothing marked: set color for whole text:  { "all" : {"fg":x,"bg":y} }
	# text marked: set color for specific text: { ..., (n,m) : {"fg":x,"bg":y} }
	# nothing marked & dict !empty: goto "nothing marked"
	# text marked & dict !empty: goto "text marked"

	# TODO: mark color setting visually via the pango Layout (set_markup)

	# TODO: determine what happens if the user edits the text...

	def _apply_color_tags(self):
		pass
	
	def _new_range_entry(self, id, before=False):
		self._ranges[id] = {"fg":None,"bg":None}
		
	def _new_attr(self, atype, colorcode, start, end):
		
		def get_gdk_color(ccolor):
			bg_color = self.get_style().base[gtk.STATE_NORMAL]
			return contrast.contrast_render_foreground_color(
					bg_color, ccolor)
		
		gcolor = get_gdk_color(colorcode)
		if atype == "fg":
			return pango.AttrForeground(gcolor.red, gcolor.green, gcolor.blue,
										start, end)
		elif atype == "bg":
			return pango.AttrBackground(gcolor.red, gcolor.green, gcolor.blue,
										start, end)
	
	def _dump_attributes(self):
		a = self.get_layout().get_attributes()
		i = a.get_iterator()
		o = []
		while True:
			l = i.get_attrs()
			o.append(l)
			if not i.next():
				break
		print o
	
	def _modify_general_color(self, fg="", bg=""):
		if not self._ranges.has_key("all"):
			self._new_range_entry("all")	
			
		elem = self._ranges["all"]
			
		if fg:
			fg_attr = self._new_attr("fg", fg, 0, self.get_text_length())
			elem["fg"] = fg_attr
			self.get_layout().get_attributes().change(fg_attr)
			self._dump_attributes()
			self.get_layout().context_changed()
		if bg:
			bg_attr = self._new_attr("bg", bg, 0, self.get_text_length())
			elem["bg"] = bg_attr
			self.get_layout().get_attributes().insert(bg_attr)
			self._dump_attributes()
			self.get_layout().context_changed()


		
	def _add_range_color(self, start, end, fg="",bg=""):
		if not bg and not fg: return
		if not self._ranges.has_key((start,end)):
			self._new_range_entry((start,end))
			
		elem = self._ranges[(start,end)]
		
		if fg:
			fg_attr = self._new_attr("fg", fg, start, end)
			elem["fg"] = fg_attr
			self.get_layout().get_attributes().insert(fg_attr)
			self._dump_attributes()
		if bg:
			bg_attr = self._new_attr("bg", bg ,start, end)
			elem["bg"] = bg_attr
			self.get_layout().get_attributes().insert(bg_attr)
			self._dump_attributes()

	def _fg_item_activate(self, item, value):
		bounds = self.get_selection_bounds()
		if not bounds:
			self._modify_general_color(fg=value)
		else:
			self._add_range_color(*bounds, fg=value)

	def _bg_item_activate(self, item, value):
		bounds = self.get_selection_bounds()
		if not bounds:
			self._modify_general_color(bg=value)
		else:
			self._add_range_color(*bounds, bg=value)

	def _reset_item_activate(self, item):
		bounds = self.get_selection_bounds()
		if not bounds:
			self._ranges = {}
		elif self._ranges.has_key((bounds[0],bounds[1])):
			del self._ranges[(bounds[0],bounds[1])]
		elif bounds:
			for (cbounds,colors) in self._ranges.items():
				if cbounds == "all": continue
				start,end = cbounds
				wstart,wend = bounds
				if wstart >= start and wend <= end:
					del self._ranges[cbounds]
					return
			
		# if we reach this, we have not found a suitable entry, maybe
		# the user means to reset it all but has a selection. reset all.
		self._ranges = {}

	def _popup_menu_handler(self,  menu):
		
		# FIXME: remove this if everything works..
		return

		fg_item = gtk.MenuItem(label=_("Foreground Color"))
		fg_submenu = gtk.Menu()

		for (value, name) in color.COLOR_NAMES.items():
			item = gtk.MenuItem(label=" ".join([n.capitalize() for n in name.split(" ")]))
			item.connect("activate", self._fg_item_activate,  value)
			fg_submenu.append(item)

		fg_item.set_submenu(fg_submenu)
		fg_item.show_all()
		menu.insert(fg_item,  0)

		bg_item = gtk.MenuItem(label=_("Background Color"))
		bg_submenu = gtk.Menu()

		for (value, name) in color.COLOR_NAMES.items():
			item = gtk.MenuItem(label=" ".join([n.capitalize() for n in name.split(" ")]))
			item.connect("activate", self._bg_item_activate,  value)
			bg_submenu.append(item)

		bg_item.set_submenu(bg_submenu)
		bg_item.show_all()
		menu.insert(bg_item,  1)
        
		reset_item = gtk.MenuItem(label=_("Reset Color"))
		reset_item.connect("activate", self._reset_item_activate)
		reset_item.show()
		menu.insert(reset_item, 2)



# coding: UTF-8
"""
Copyright (c) 2009 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

from typecheck import types
from messages import print_notification, print_error, print_normal
from gettext import gettext as _
# TODO: use gettext

import tabs
import config
import signals
import connection
import plugin_control

def no_connection():
	if not connection.sushi.connected:
		print_error("No connection to maki.")
		return True
	return False

def send_text(tab, text):
	if no_connection() or type(tab) == tabs.Server or not text:
		return

	server_name = tab.parent.name

	if not tab.connected:
		print_error("Not connected to server "\
			"'%s'." % (server_name))
		return

	if type(tab) == tabs.Channel and not tab.joined:
		print_error("You are not on the channel "\
			"'%s'." % (tab.name))
		return

	connection.sushi.message(server_name, tab.name, text)

def need_active_tab(fun):
	""" decorator for command handlers which require an active tab """
	def notifier(main_window, argv):
		if not main_window.current_tab:
			return print_error("You have to be on an active tab.")
		return fun(main_window, argv)
	return notifier

def need_sushi(fun):
	""" decorator for command handlers which require a sushi conn. """
	def notifier(main_window, argv):
		if no_connection():
			return
		return fun(main_window, argv)
	return notifier

# Command handler

@need_sushi
def cmd_add_server(main_window, argv):
	""" /add_server <name> <address> <port> <nick> <real> """
	if no_connection():
		return

	if len(argv) != 6:
		print_notification("Usage: /add_server <name> <address> "\
			"<port> <nick> <real>")
		return

	cmd, server_name, address, port, nick, name = argv

	try:
		int(port)
	except ValueError:
		print_error("Invalid value for 'port' given.")
		return

	if server_name in connection.sushi.server_list("",""):
		print_error("Server with name '%s' already existing." % (server_name))
		return

	pairs = (("address", address), ("port", port), ("nick", nick),
		("name", name))
	for key,value in pairs:
		connection.sushi.server_set(server_name, "server", key, value)

	print_notification("Server '%s' successfully added." % (server_name))

@need_sushi
@need_active_tab
def cmd_away(main_window, argv):
	""" /away <message> """
	# TODO default away message
	if len(argv) < 2:
		return print_notification("/away <message>")
	server = tabs.get_server(main_window.current_tab)
	connection.sushi.away(server.name, " ".join(argv[1:]))

@need_sushi
@need_active_tab
def cmd_back(main_window, argv):
	""" /back """
	server = tabs.get_server(main_window.current_tab)
	connection.sushi.back(server.name)

def cmd_clear(main_window, argv):
	""" /clear [<markup>] """
	if len(argv) == 2 and argv[1] == "markup":
		for tab in tabs.tree_to_list(main_window.servers):
			tab.reset_status()
		main_window.update_divider()
	else:
		del main_window.body.body[0:]

def cmd_close(main_window, argv):
	""" /close """
	# close the current tab
	if not main_window.current_tab:
		return

	to_delete = main_window.current_tab
	main_window.switch_to_next_tab()
	to_delete.remove()

@need_sushi
def cmd_connect(main_window, argv):
	""" /connect <server> """
	if no_connection():
		return

	if len(argv) != 2:
		print_notification("Usage: /connect <server>")
		return

	if argv[1] not in connection.sushi.server_list("",""):
		print_error("connect: The server '%(server)s' is not known." % {
			"server": argv[1]})
		return

	connection.sushi.connect(argv[1])

@need_sushi
@need_active_tab
def cmd_dcc(main_window, argv):
	""" /dcc chat <target>
	/dcc send <target> <path>
	/dcc list <type>
	/dcc remove <type> <id>
	"""
	if no_connection():
		return

	if len(argv) < 2:
		print_notification("Usage: /dcc accept|chat|send|list|remove")
		return

	server_tab = tabs.get_server(main_window.current_tab)

	if argv[1] == "send":
		if len(argv) < 4:
			print_notification("Usage: /dcc send <target> <path>")
			return

		connection.sushi.dcc_send(server_tab.name, argv[2], argv[3])
	elif argv[1] == "accept":
		def accepted_cb(time, id, *x):
			if id == int(argv[2]):
				print_notification(_("DCC action with id %(id)d accepted.") % {
					"id": id})
				signals.disconnect_signal("dcc_send", accepted_cb)

		if len(argv) < 3:
			print_notification("Usage: /dcc accept <id>")
			return

		signals.connect_signal("dcc_send", accepted_cb)
		connection.sushi.dcc_send_accept(int(argv[2]))
	elif argv[1] == "chat":
		pass
	elif argv[1] == "list":
		if len(argv) < 3:
			print_notification("Usage: /dcc list chat|send")
			return

		if argv[2] == "send":
			ids, servers, froms, filenames, sizes, progresses, speeds, statuses = connection.sushi.dcc_sends()

			for i in range(len(ids)):
				print_normal("#%(id)d: %(filename)s from/to %(from)s [%(server)s]: %(progress)d/%(size)d @ %(speed)d" % {
						"id": ids[i],
						"filename": filenames[i],
						"from": froms[i],
						"server": servers[i],
						"progress": progresses[i],
						"size": sizes[i],
						"speed": speeds[i]
					})
		elif argv[2] == "chat":
			pass
	elif argv[1] == "remove":
		if len(argv) < 4:
			print_notification("Usage: /dcc remove <type> <id>")
			return

		if argv[2] == "send":
			connection.sushi.dcc_send_remove(int(argv[3]))
		elif argv[2] == "chat":
			pass

def cmd_echo(main_window, argv):
	""" /echo <text> """
	main_window.print_text(" ".join(argv[1:])+"\n")

def cmd_help(main_window, argv):
	""" /help [<command>] """
	if len(argv) == 2:
		sCmd = argv[1]
	else:
		sCmd = None

	msg = ""

	for (name, value) in globals().items():
		cmd_name = name[4:]

		if name[0:4] == "cmd_":
			if sCmd and sCmd != cmd_name:
				continue

			msg += "%s:\n %s\n" % (cmd_name,
				(value.__doc__
				or "No documentation yet.").replace("\t", "   "))

			if sCmd and sCmd == cmd_name:
				break

	print_normal(msg)

@need_sushi
@need_active_tab
def cmd_ignore(main_window, argv):
	""" /ignore <pattern> """
	if len(argv) != 2:
		return print_notification("Usage: /ignore <pattern>")

	server = tabs.get_server(main_window.current_tab).name
	connection.sushi.ignore(server, argv[1])

	print_notification("Ignored user %(user)s." % {"user":argv[1]})

@need_sushi
@need_active_tab
def cmd_ignores(main_window, argv):
	""" /ignores """
	server = tabs.get_server(main_window.current_tab).name
	ignores = connection.sushi.ignores(server)

	if not ignores:
		print_normal("No users ignored on %(server)s." % {"server":server})
	else:
		print_normal("Ignored users: %(ulist)s." % {
			"ulist":", ".join(ignores)})

@need_sushi
@need_active_tab
def cmd_join(main_window, argv):
	""" /join <channel> [<key>] """
	key = ""

	if len(argv) == 3:
		cmd, channel, key = argv
	elif len(argv) == 2:
		cmd, channel = argv
	elif len(argv) == 1 and type(main_window.current_tab) == tabs.Channel:
		channel = main_window.current_tab.name
	else:
		print_notification("Usage: /join [<channel> [<key>]]")
		return

	server_tab = tabs.get_server(main_window.current_tab)

	connection.sushi.join(server_tab.name, channel, key)
	print_notification("Joining channel '%s'." % (channel))

@need_sushi
@need_active_tab
def cmd_kick(main_window, argv):
	""" /kick <user> [<reason>]
		Kick a user from the current channel with an optional reason.
	"""
	if len(argv) < 2:
		return print_notification("Usage: /kick <user> [<reason>]")
	ct = main_window.current_tab
	if type(ct) != tabs.Channel:
		return print_error("You have to be on a channel to do so.")
	connection.sushi.kick(ct.parent.name, ct.name, argv[1],
		" ".join(argv[2:]))

def cmd_load(main_window, argv):
	if len(argv) != 2:
		print_notification("Usage: /load <filename>")
		return

	name = argv[1]

	if plugin_control.is_loaded(name):
		print_notification("Plugin '%s' already loaded." % (name))
		return

	if plugin_control.load(name):
		print_notification("Plugin '%s' successfully loaded." % (name))
	else:
		print_notification("Error while loading plugin '%s'." % (name))

def cmd_maki(main_window, argv):
	""" /maki [connect|shutdown] """
	def usage():
		print_notification("Usage: /maki [connect|disconnect|shutdown]")

	if len(argv) != 2:
		usage()
		return

	cmd = argv[1]

	if cmd == "connect":
		print_notification("Connection to maki...")

		if connection.connect():
			print_notification("Connection to maki etablished.")
		else:
			print_notification("Connection to maki failed.")
		main_window.update_divider()

	elif cmd == "disconnect":
		if no_connection():
			return

		connection.disconnect()
		print_notification("Disconnected from maki.")
		main_window.update_divider()

	elif cmd == "shutdown":
		if no_connection():
			return

		connection.sushi.shutdown(config.get("chatting", "quit_message"))
		main_window.update_divider()

	else:
		usage()
		return

@need_sushi
def cmd_me(main_window, argv):
	""" /me <text>
		Action in third person.
	"""
	ct = main_window.current_tab
	if not type(ct) in (tabs.Channel, tabs.Query):
		return print_notification("You have to be on a channel or query.")
	connection.sushi.action(ct.parent.name, ct.name, " ".join(argv[1:]))

@need_sushi
@need_active_tab
def cmd_mode(main_window, argv):
	""" /mode <target> [[+|-]<mode> [<param>]]
		Set (+|-) or query a mode on a target with an optional parameter.
		If wether mode nor param are given, this command queries the
		current modes set.
	"""
	if len(argv) < 2:
		return print_notification(
			"Usage: /mode <target> [[+|-]<mode> [<param>]]")

	server = tabs.get_server(main_window.current_tab).name

	connection.sushi.mode(server, argv[1], " ".join(argv[2:]))

@need_sushi
@need_active_tab
def cmd_names(main_window, argv):
	""" /names """
	ct = main_window.current_tab
	if type(ct) != tabs.Channel:
		print_error("You must be on a channel to use /names.")
		return
	connection.sushi.names(ct.parent.name, ct.name)

@need_sushi
@need_active_tab
def cmd_nick(main_window, argv):
	""" /nick <new nick>
		Changes the current nick to a new one.
	"""
	if len(argv) != 2:
		return print_notification("Usage: /nick <new nick>")

	server = tabs.get_server(main_window.current_tab)

	connection.sushi.nick(server.name, argv[1])

@need_sushi
@need_active_tab
def cmd_notice(main_window, argv):
	""" /notice <user> <message>
		Send a notice to a user.
	"""
	if len(argv) < 3:
		return print_notification("Usage: /notice <user> <message>")
	server = tabs.get_server(main_window.current_tab)
	connection.sushi.notice(server.name, argv[1], " ".join(argv[2]))

def cmd_overview(main_window, argv):
	""" /overview [<server>] """
	if no_connection():
		return

	if len(argv) == 2:
		if argv[1] not in [n.name for n in main_window.servers]:
			print_error("Server '%s' not found." % (argv[1]))
			return
		server_list = [main_window.find_server(argv[1])]
	else:
		server_list = main_window.servers

	print_normal("Begin overview.")

	for server in server_list:
		if server.connected:
			connected = "connected"
		else:
			connected = "not connected"

		print_normal("Server '%s' (%s)" % (server, connected))

		for child in server.children:
			if type(child) == tabs.Channel:
				if child.joined:
					joined = "joined"
				else:
					joined = "not joined"
				print_normal("- Channel '%s' (%s)" % (child.name, joined))
			else:
				print_normal("- Query '%s'" % (child.name))

	print_normal("End of overview.")

@need_sushi
def cmd_part(main_window, argv):
	""" /part [<reason>] """
	if no_connection():
		return

	if len(argv) == 2:
		reason = " ".join(argv[1:])
	else:
		reason = ""

	ct = main_window.current_tab
	if type(ct) in (tabs.Server, tabs.Query):
		print_error("The current tab is not a channel.")
		return
	else:
		connection.sushi.part(ct.parent.name, ct.name, reason)

def cmd_python(main_window, argv):
	""" /python <code> """
	print_notification ("python: Executing...\n")
	try:
		exec(" ".join(argv[1:]))
	except BaseException,e:
		print_error ("python: %s\n" % (e))

def cmd_quit(main_window, argv):
	""" /quit """
	main_window.quit()

def cmd_reload(main_window, argv):
	""" /reload
		Reload the configuration.
	"""
	# FIXME
	config.setup()
	print_notification("Config reloaded.")

@need_sushi
def cmd_remove_server(main_window, argv):
	""" /remove_server <name> """
	if no_connection():
		return

	if len(argv) != 2:
		print_notification("Usage: /remove_server <name>")
		return

	name = argv[1]

	if not name in connection.sushi.server_list("", ""):
		print_error("Server '%s' not known to maki." % (name))
		return

	connection.sushi.server_remove(name, "", "")
	print_notification("Server '%s' successfully removed." % (name))

@need_sushi
def cmd_servers(main_window, argv):
	""" /servers """
	active_servers = connection.sushi.servers()

	print_normal("Servers known to maki: ")
	for name in connection.sushi.server_list("",""):
		if name in active_servers:
			connected = "connected"
		else:
			connected = "not connected"
		print_normal("- '%s' (%s)" % (name, connected))

@need_sushi
@need_active_tab
def cmd_topic(main_window, argv):
	""" /topic [<new topic>] """
	ct = main_window.current_tab
	if type(ct) != tabs.Channel:
		return print_error("You have to be on a channel to do so.")
	connection.sushi.topic(ct.parent.name, ct.name, " ".join(argv[1:]))

@need_sushi
@need_active_tab
def cmd_unignore(main_window, argv):
	if len(argv) != 2:
		return print_notification("Usage: /unignore <pattern>")

	server = tabs.get_server(main_window.current_tab).name
	connection.sushi.unignore(server, argv[1])

	print_notification("Unignored user %(user)s." % {"user":argv[1]})

def cmd_unload(main_window, argv):
	if len(argv) != 2:
		print_notification("Usage: /unload <filename>")
		return
	name = argv[1]

	if not plugin_control.is_loaded(name):
		print_error("Plugin '%s' is not loaded." % (name))
		return

	if not plugin_control.unload(name):
		print_error("Failed to unload plugin '%s'." % (name))
	else:
		print_notification("Plugin '%s' successfully unloaded." % (name))

_command_dict = {
	"add_server": cmd_add_server,
	"away" : cmd_away,
	"back" : cmd_back,
	"clear": cmd_clear,
	"close": cmd_close,
	"connect": cmd_connect,
	"dcc": cmd_dcc,
	"echo": cmd_echo,
	"help": cmd_help,
	"ignore" : cmd_ignore,
	"ignores" : cmd_ignores,
	"j": cmd_join,
	"join": cmd_join,
	"kick" : cmd_kick,
	"load": cmd_load,
	"maki": cmd_maki,
	"me" : cmd_me,
	"mode" : cmd_mode,
	"names": cmd_names,
	"nick" : cmd_nick,
	"notice" : cmd_notice,
	"overview": cmd_overview,
	"part": cmd_part,
	"python": cmd_python,
	"quit": cmd_quit,
	"reload" : cmd_reload,
	"remove_server": cmd_remove_server,
	"servers": cmd_servers,
	"topic" : cmd_topic,
	"unignore" : cmd_unignore,
	"unload": cmd_unload
}

_builtins = _command_dict.keys()

def strip_command_char (cmd):
	if not cmd: return cmd
	if cmd[0] != config.get("nigiri","command_char"):
		return cmd
	return cmd[1:]

def parse(main_window, text):
	argv = text.split(" ")

	if (not argv
		or (argv[0]
			and argv[0][0] != config.get(
				"nigiri",
				"command_char"))):
		return False

	command = strip_command_char (argv[0])

	if not command:
		return False

	try:
		fun = _command_dict[command]
	except KeyError:
		print_error ("No such command '%s'.\n" % (command))
	else:
		fun (main_window, argv)

	return True

@types (name=basestring, fun=type(lambda:""))
def add_command(name, fun):
	global _command_dict
	if _command_dict.has_key (name):
		return False
	_command_dict[name] = fun
	return True

@types (name=basestring)
def remove_command(name):
	global _command_dict
	if _command_dict.has_key (name) and name not in _builtins:
		del _command_dict[name]
		return True
	return False



# coding:utf-8
"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import gobject
import pango
import gettext
from gettext import gettext as _
from threading import Timer

from ... import config

from ..builder import build_status_icon, error_dialog
from .._builder import widgets

from ...helper import color
from ...helper import code
from ...typecheck import types

from ...lib.inline_dialog import InlineMessageDialog
from ...lib.welcome_window import WelcomeWindow

from . import visibility

def get_font():

	if not config.get_bool("tekka", "use_default_font"):
		return config.get("tekka", "font")

	try:
		import gconf

		client = gconf.client_get_default()

		font = client.get_string(
			"/desktop/gnome/interface/monospace_font_name")

		return font

	except:
		return config.get("tekka", "font")


def apply_new_font():
	""" iterate over all widgets which use fonts and change them """

	font = get_font()

	for row in widgets.get_object("tabs_view").get_model():
		for child in row.iterchildren():
			set_font(child[0].window.textview, font)
		set_font(row[0].window.textview, font)

	set_font(widgets.get_object("output"), font)
	set_font(widgets.get_object("input_entry"), font)
	set_font(widgets.get_object("general_output"), font)


@types(switch=bool)
def set_useable(switch):
	"""
		Dis- or enable the widgets
		which emit or receive signals
		to/from maki.
	"""
	global gui_is_useable

	widgetList = [
		widgets.get_object("input_entry"),
		widgets.get_object("tabs_view"),
		widgets.get_object("nicks_view"),
		widgets.get_object("general_output_window")
	]

	if switch or not is_welcome_screen():
		# don't disable welcome screen, it knows better
		widgetList.append(widgets.get_object("output"))
		widgetList.append(widgets.get_object("output_shell"))

	for widget in widgetList:
		widget.set_sensitive(switch)

	if switch: widgets.get_object("input_entry").grab_focus()

	gui_is_useable = switch


def is_welcome_screen():
	""" returns True if the welcome screen is displayed """
	return type(widgets.get_object("output_window")) == WelcomeWindow


def has_focus():
	""" return wether the mainwindow has focus or not """

	win = widgets.get_object("main_window")

	return win.has_toplevel_focus()


@types(switch=bool)
def set_urgent(switch):
	""" Sets or unsets the urgent status to the main window.
		If the status icon is enabled it will be set flashing.
	"""
	win = widgets.get_object("main_window")

	if has_focus():
		# don't urgent if we have already the focus
		return

	win.set_urgency_hint(switch)

	statusIcon = widgets.get_object("status_icon")

	if statusIcon:
		statusIcon.set_blinking(switch)


@types(title=basestring)
def set_window_title(title):
	""" Sets the window title to the main window. """
	widgets.get_object("main_window").set_title(title)


@types(nick=basestring)
def set_nick(nick):
	""" Sets nick as label text of nick_label. """
	widgets.get_object("nick_label").set_text(nick)


@types(normal=int, ops=int)
def set_user_count(normal, ops):
	""" sets the amount of users in the current channel. """

	m_users = gettext.ngettext(
		"%d User", "%d Users", normal) % (normal)
	m_ops = gettext.ngettext(
		"%d Operator", "%d Operators", ops) % (ops)

	widgets.get_object("nick_stats_label").set_text(
		"%(users)s – %(ops)s" % {
			"users": m_users, "ops": m_ops })


def set_font(textView, font):
	"""	Sets the font of the textView to
		the font identified by fontFamily
	"""
	fd = pango.FontDescription(font)

	if not fd:
		logging.error("set_font: Font _not_ modified (previous error)")

	else:
		textView.modify_font(fd)


@types(string=basestring)
def set_topic(string):
	""" Sets the given string as text in
		the topic bar.
	"""
	tb = widgets.get_object("topic_label")
	tb.set_markup(string)
	tb.set_tooltip_markup(string)


def clear_all_outputs():

	from .. import tabs

	def clear(buf):
		if buf: buf.set_text("")

	current_tab = tabs.get_current_tab()

	if current_tab:
		output = current_tab.window.textview

		buf = output.get_buffer()

		clear(buf)

	buf = widgets.get_object("general_output").get_buffer()

	clear(buf)


@types(string=basestring, html=bool)
def myPrint(string, html=False):
	"""
		prints the string `string` in the current output
		buffer. If html is true the string would be inserted via
		the insert_html-method falling back to normal insert
		if it's not possible to insert via insert_html.
	"""
	textview = widgets.get_object("output")
	output = textview.get_buffer()

	if not output:
		logging.error("myPrint: No output buffer.")
		return

	if not html:

		if output.get_char_count() > 0:

			string = "\n" + string

		output.insert(output.get_end_iter(), string)

	else:

		try:

			output.insert_html(output.get_end_iter(), string)

		except AttributeError:

			logging.info("myPrint: No HTML buffer, printing normal.")
			output.insert(output.get_end_iter(), "\n"+string)

	textview.scroll_to_bottom()


def show_error_dialog(title = "", message = ""):
	""" create a dialog with error_dialog() and show  it up.
		The dialog closes on every action.
	"""
	d = error_dialog(title, message)

	d.connect("response", lambda d,i: d.destroy())
	d.show()

	return d


def show_maki_connection_error(title, message):
	d = InlineMessageDialog(
		_("tekka could not connect to maki."),
		_("Please check whether maki is running."))

	d.connect("response", lambda d,id: d.destroy())

	show_inline_dialog(d)


def show_inline_message(title, message, dtype="info"):
	""" title, the title of the dialog
		message, the message
		dtype, the type. Values can be "error","info" or "warning"
	"""
	if dtype == "error":
		icon = gtk.STOCK_DIALOG_ERROR
	elif dtype == "info":
		icon = gtk.STOCK_DIALOG_INFO
	elif dtype == "warning":
		icon = gtk.STOCK_DIALOG_WARNING

	d = InlineMessageDialog(title, message, icon=icon)
	d.connect("response", lambda d,id: d.destroy())

	show_inline_dialog(d)

	return d


def show_inline_dialog(dialog):
	""" show an InlineDialog in the notification_vbox"""

	# Purpose: auto removing messages (depends on config)
	self = code.init_function_attrs(
				show_inline_dialog,
				timeouts = [])

	area = widgets.get_object("notification_vbox")

	if dialog:

		area.set_no_show_all(False)
		area.add(dialog)
		area.show_all()
		area.set_no_show_all(True)

		if config.get_bool("tekka", "idialog_timeout"):

			def dialog_timeout_cb():
				area.remove(dialog)
				self.timeouts.remove(dialog_timeout_cb.timer)

			if isinstance(dialog, InlineMessageDialog):
				# let only messages disappear
				t = Timer(
					int(config.get("tekka", "idialog_timeout_seconds")),
					dialog_timeout_cb)

				dialog_timeout_cb.timer = t
				self.timeouts.append(t)

				t.start()

	else:
		area.set_property("visible", False)

		for timer in self.timeouts:
			t.cancel()
